# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for point."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.representation import point
from tensorflow_graphics.util import test_case


class PointTest(test_case.TestCase):

  @parameterized.parameters(
      ((None,), (None,), (None,)),
      ((None, None), (None, None), (None, None)),
      ((3,), (3,), (3,)),
      ((2, 3), (2, 3), (2, 3)),
      ((1, 2, 3), (2, 2, 3), (2, 3)),
  )
  def test_distance_to_ray_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(point.distance_to_ray, shapes)

  @parameterized.parameters(
      ("must have the same number of dimensions", (0,), (None,), (None,)),
      ("must have the same number of dimensions", (None,), (0,), (None,)),
      ("must have the same number of dimensions", (None,), (None,), (0,)),
      ("must have the same number of dimensions", (0,), (1,), (1,)),
      ("must have the same number of dimensions", (2,), (1,), (2,)),
      ("must have the same number of dimensions", (2, 3), (2, 2), (2, 0)),
      ("Not all batch dimensions are broadcast-compatible.", (0, 3), (3, 3),
       (3, 3)),
      ("Not all batch dimensions are broadcast-compatible.", (1, 3), (2, 3),
       (3, 3)),
  )
  def test_distance_to_ray_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(point.distance_to_ray, error_msg, shapes)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_distance_to_ray_jacobian_random(self):
    """Tests the Jacobian of the distance to ray function."""
    eps = sys.float_info.epsilon
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    point_init = np.random.random(size=tensor_shape + [3])
    origin_init = np.random.random(size=tensor_shape + [3])
    direction_init = np.random.random(size=tensor_shape + [3])
    direction_init /= np.maximum(
        np.linalg.norm(direction_init, axis=-1, keepdims=True), eps)

    self.assert_jacobian_is_correct_fn(
        lambda x: point.distance_to_ray(x, origin_init, direction_init),
        [point_init])
    self.assert_jacobian_is_correct_fn(
        lambda x: point.distance_to_ray(point_init, x, direction_init),
        [origin_init])
    self.assert_jacobian_is_correct_fn(
        lambda x: point.distance_to_ray(point_init, origin_init, x),
        [direction_init])

  @parameterized.parameters(
      (((1., 1., 1.), (-10., 1., 1.), (1., 0., 0.)), ((0.,),)),)
  def test_distance_to_ray_preset(self, test_inputs, test_outputs):
    """Tests 0 distance from point to ray."""
    self.assert_output_is_correct(point.distance_to_ray, test_inputs,
                                  test_outputs)

  def test_distance_to_ray_random(self):
    """Tests distance point to ray."""
    eps = sys.float_info.epsilon
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    random_origin = np.random.random(size=tensor_shape + [3])
    random_direction = np.random.random(size=tensor_shape + [3])
    random_direction /= np.maximum(
        np.linalg.norm(random_direction, axis=-1, keepdims=True), eps)
    random_distances = np.random.random(size=tensor_shape + [1]) * 1000
    x, y, _, _ = np.split(random_direction, (1, 2, 3), axis=-1)

    # Find perpendicular vector.
    perp = np.concatenate((-y, x, np.zeros_like(x)), axis=-1)
    perp /= np.maximum(np.linalg.norm(perp, axis=-1, keepdims=True), eps)
    # Choose a point on perpendicular unit vector at the good distance.
    points = random_origin + perp * random_distances
    distance = point.distance_to_ray(points, random_origin, random_direction)

    self.assertAllClose(random_distances, distance)

  @parameterized.parameters(
      ((None,), (None,), (None,)),
      ((None, None), (None, None), (None, None)),
      ((3,), (3,), (3,)),
      ((2, 3), (2, 3), (2, 3)),
      ((1, 2, 3), (2, 2, 3), (2, 3)),
  )
  def test_project_to_ray_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(point.project_to_ray, shapes)

  @parameterized.parameters(
      ("must have the same number of dimensions", (0,), (None,), (None,)),
      ("must have the same number of dimensions", (None,), (0,), (None,)),
      ("must have the same number of dimensions", (None,), (None,), (0,)),
      ("must have the same number of dimensions", (0,), (1,), (1,)),
      ("must have the same number of dimensions", (2,), (1,), (2,)),
      ("must have the same number of dimensions", (2, 3), (2, 2), (2, 0)),
      ("Not all batch dimensions are broadcast-compatible.", (0, 3), (3, 3),
       (3, 3)),
      ("Not all batch dimensions are broadcast-compatible.", (1, 3), (2, 3),
       (3, 3)),
  )
  def test_project_to_ray_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(point.project_to_ray, error_msg, shapes)

  @parameterized.parameters(
      (((1., 1., 1.), (-10., 1., 1.), (1., 0., 0.)), ((1., 1., 1.),)),)
  def test_project_to_ray_preset(self, test_inputs, test_outputs):
    """Tests that a point on a ray projects to itself."""
    self.assert_output_is_correct(point.project_to_ray, test_inputs,
                                  test_outputs)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_project_to_ray_jacobian_random(self):
    """Tests the Jacobian of the distance to ray function."""
    eps = sys.float_info.epsilon
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    point_init = np.random.random(size=tensor_shape + [3])
    origin_init = np.random.random(size=tensor_shape + [3])
    direction_init = np.random.random(size=tensor_shape + [3])
    direction_init /= np.maximum(
        np.linalg.norm(direction_init, axis=-1, keepdims=True), eps)
    point_tensor = tf.convert_to_tensor(value=point_init)
    origin_tensor = tf.convert_to_tensor(value=origin_init)
    direction_tensor = tf.convert_to_tensor(value=direction_init)

    self.assert_jacobian_is_correct_fn(
        lambda x: point.project_to_ray(x, origin_tensor, direction_tensor),
        [point_init])
    self.assert_jacobian_is_correct_fn(
        lambda x: point.project_to_ray(point_tensor, x, direction_tensor),
        [origin_init])
    self.assert_jacobian_is_correct_fn(
        lambda x: point.project_to_ray(point_tensor, origin_tensor, x),
        [direction_init])

  def test_project_to_ray_random(self):
    """Tests the function that projects a point to a ray."""
    eps = sys.float_info.epsilon
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    random_origin = np.random.random(size=tensor_shape + [3])
    random_direction = np.random.random(size=tensor_shape + [3])
    random_direction /= np.maximum(
        np.linalg.norm(random_direction, axis=-1, keepdims=True), eps)
    random_distances = np.random.random(size=tensor_shape + [1]) * 1000
    x, y, _, _ = np.split(random_direction, (1, 2, 3), axis=-1)

    # Find perpendicular vector.
    perp = np.concatenate((-y, x, np.zeros_like(x)), axis=-1)
    perp /= np.maximum(np.linalg.norm(perp, axis=-1, keepdims=True), eps)
    # Choose a point on perpendicular unit vector at the good distance.
    points = random_origin + perp * random_distances
    points = point.project_to_ray(points, random_origin, random_direction)

    self.assertAllClose(random_origin, points)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Tests for ray."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from absl import flags
from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.representation import ray
from tensorflow_graphics.util import test_case

FLAGS = flags.FLAGS


class RayTest(test_case.TestCase):

  def _generate_random_example(self):
    num_cameras = 4
    num_keypoints = 3
    batch_size = 2
    self.points_values = np.random.random_sample((batch_size, num_keypoints, 3))
    points_expanded_values = np.expand_dims(self.points_values, axis=-2)
    startpoints_values = np.random.random_sample(
        (batch_size, num_keypoints, num_cameras, 3))

    difference = points_expanded_values - startpoints_values
    difference_norm = np.sqrt((difference * difference).sum(axis=-1))
    direction = difference / np.expand_dims(difference_norm, axis=-1)

    self.startpoints_values = points_expanded_values - 0.5 * direction
    self.endpoints_values = points_expanded_values + 0.5 * direction
    self.weights_values = np.ones((batch_size, num_keypoints, num_cameras))

    # Wrap these with identies because some assert_* ops look at the constant
    # tensor values and mark these as unfeedable.
    self.points = tf.identity(tf.convert_to_tensor(value=self.points_values))
    self.startpoints = tf.identity(
        tf.convert_to_tensor(value=self.startpoints_values))
    self.endpoints = tf.identity(
        tf.convert_to_tensor(value=self.endpoints_values))
    self.weights = tf.identity(tf.convert_to_tensor(value=self.weights_values))

  @parameterized.parameters(
      ("Not all batch dimensions are identical.", (4, 3), (5, 3), (4,)),
      ("must have exactly 3 dimensions in axis", (4, 2), (4, 2), (4,)),
      ("must have a rank greater than 1", (3,), (3,), (None,)),
      ("must have greater than 1 dimensions in axis -2", (1, 3), (1, 3), (1,)),
      ("Not all batch dimensions are identical.", (2, 4, 3), (2, 4, 3), (2, 5)),
  )
  def test_triangulate_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(ray.triangulate, error_msg, shapes)

  @parameterized.parameters(
      ((4, 3), (4, 3), (4,)),
      ((5, 4, 3), (5, 4, 3), (5, 4)),
      ((6, 5, 4, 3), (6, 5, 4, 3), (6, 5, 4)),
  )
  def test_triangulate_exception_is_not_raised(self, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_not_raised(ray.triangulate, shapes)

  def test_triangulate_jacobian_is_correct(self):
    """Tests that Jacobian is correct."""
    self._generate_random_example()

    self.assert_jacobian_is_correct_fn(
        lambda x: ray.triangulate(x, self.endpoints, self.weights),
        [self.startpoints_values])
    self.assert_jacobian_is_correct_fn(
        lambda x: ray.triangulate(self.startpoints, x, self.weights),
        [self.endpoints_values])
    self.assert_jacobian_is_correct_fn(
        lambda x: ray.triangulate(self.startpoints, self.endpoints, x),
        [self.weights_values])

  def test_triangulate_jacobian_is_finite(self):
    """Tests that Jacobian is finite."""
    self._generate_random_example()

    self.assert_jacobian_is_finite_fn(
        lambda x: ray.triangulate(x, self.endpoints, self.weights),
        [self.startpoints_values])
    self.assert_jacobian_is_finite_fn(
        lambda x: ray.triangulate(self.startpoints, x, self.weights),
        [self.endpoints_values])
    self.assert_jacobian_is_finite_fn(
        lambda x: ray.triangulate(self.startpoints, self.endpoints, x),
        [self.weights_values])

  def test_triangulate_random(self):
    """Tests that original points are recovered by triangualtion."""
    self._generate_random_example()
    test_inputs = (self.startpoints, self.endpoints, self.weights)
    test_outputs = (self.points_values,)

    self.assert_output_is_correct(
        ray.triangulate,
        test_inputs,
        test_outputs,
        rtol=1e-05,
        atol=1e-08,
        tile=False)

  def test_negative_weights_exception_raised(self):
    """Tests that exceptions are properly raised."""
    self._generate_random_example()
    self.weights = -1.0 * tf.ones_like(self.weights, dtype=tf.float64)

    with self.assertRaises(tf.errors.InvalidArgumentError):
      points = ray.triangulate(self.startpoints, self.endpoints, self.weights)
      self.evaluate(points)

  def test_less_that_two_nonzero_weights_exception_raised(self):
    """Tests that exceptions are properly raised."""
    self._generate_random_example()
    self.weights = tf.convert_to_tensor(
        value=np.array([[[1., 1., 0., 0.], [1., 1., 0., 0.], [1., 1., 0., 0.]],
                        [[1., 1., 0., 0.], [1., 1., 0., 0.], [1., 0., 0., 0.]]],
                       dtype=np.float64))

    with self.assertRaises(tf.errors.InvalidArgumentError):
      points = ray.triangulate(self.startpoints, self.endpoints, self.weights)
      self.evaluate(points)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis 0", (2,), (1,), (3,), (3,)),
      ("must have a rank of 1", (2, 3), (1,), (3,), (3,)),
      ("must have exactly 1 dimensions in axis 0", (3,), (2,), (3,), (3,)),
      ("must have a rank of 1", (3,), (2, 1), (3,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (1,), (2,), (3,)),
      ("must have exactly 3 dimensions in axis -1", (3,), (1,), (3,), (2,)),
      ("Not all batch dimensions are identical.", (3,), (1,), (3,), (2, 3)),
  )
  def test_intersection_ray_sphere_shape_raised(self, error_msg, *shapes):
    """tests that exceptions are raised when shapes are not supported."""
    self.assert_exception_is_raised(ray.intersection_ray_sphere, error_msg,
                                    shapes)

  @parameterized.parameters(
      ((3,), (1,), (3,), (3,)),
      ((3), (1), (None, 3), (None, 3)),
  )
  def test_intersection_ray_sphere_shape_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised on supported shapes."""
    self.assert_exception_is_not_raised(ray.intersection_ray_sphere, shapes)

  def test_intersection_ray_sphere_exception_raised(self):
    """Tests that exceptions are properly raised."""
    sphere_center = np.random.uniform(size=(3,))
    point_on_ray = np.random.uniform(size=(3,))
    sample_ray = np.random.uniform(size=(3,))
    normalized_sample_ray = sample_ray / np.linalg.norm(sample_ray, axis=-1)
    positive_sphere_radius = np.random.uniform(
        sys.float_info.epsilon, 1.0, size=(1,))
    negative_sphere_radius = np.random.uniform(-1.0, 0.0, size=(1,))

    with self.subTest(name="positive_radius"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            ray.intersection_ray_sphere(sphere_center, negative_sphere_radius,
                                        normalized_sample_ray, point_on_ray))

    with self.subTest(name="normalized_ray"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            ray.intersection_ray_sphere(sphere_center, positive_sphere_radius,
                                        sample_ray, point_on_ray))

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_intersection_ray_sphere_jacobian_random(self):
    """Test the Jacobian of the intersection_ray_sphere function."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    sphere_center_init = np.random.uniform(0.0, 1.0, size=(3,))
    sphere_radius_init = np.random.uniform(10.0, 11.0, size=(1,))
    ray_init = np.random.uniform(size=tensor_shape + [3])
    ray_init /= np.linalg.norm(ray_init, axis=-1, keepdims=True)
    point_on_ray_init = np.random.uniform(0.0, 1.0, size=tensor_shape + [3])

    def intersection_ray_sphere_position(sphere_center, sphere_radius,
                                         input_ray, point_on_ray):
      y_p, _ = ray.intersection_ray_sphere(sphere_center, sphere_radius,
                                           input_ray, point_on_ray)
      return y_p

    def intersection_ray_sphere_normal(sphere_center, sphere_radius, input_ray,
                                       point_on_ray):
      _, y_n = ray.intersection_ray_sphere(sphere_center, sphere_radius,
                                           input_ray, point_on_ray)
      return y_n

    self.assert_jacobian_is_correct_fn(
        intersection_ray_sphere_position,
        [sphere_center_init, sphere_radius_init, ray_init, point_on_ray_init])
    self.assert_jacobian_is_correct_fn(
        intersection_ray_sphere_normal,
        [sphere_center_init, sphere_radius_init, ray_init, point_on_ray_init])

  @parameterized.parameters(
      (((0.0, 0.0, 3.0), (1.0,), (0.0, 0.0, 1.0), (0.0, 0.0, 0.0)),
       (((0.0, 0.0, 2.0), (0.0, 0.0, 4.0)), ((0.0, 0.0, -1.0),
                                             (0.0, 0.0, 1.0)))),
      (((0.0, 0.0, 3.0), (1.0,), (0.0, 0.0, 1.0), (1.0, 0.0, 0.0)),
       (((1.0, 0.0, 3.0), (1.0, 0.0, 3.0)), ((1.0, 0.0, 0.0),
                                             (1.0, 0.0, 0.0)))),
  )
  def test_intersection_ray_sphere_preset(self, test_inputs, test_outputs):
    self.assert_output_is_correct(
        ray.intersection_ray_sphere, test_inputs, test_outputs, tile=False)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains routines shared for rotation matrices."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def is_valid(matrix, atol=1e-3, name=None):
  r"""Determines if a matrix in K-dimensions is a valid rotation matrix.

  Determines if a matrix $$\mathbf{R}$$ is a valid rotation matrix by checking
  that $$\mathbf{R}^T\mathbf{R} = \mathbf{I}$$ and $$\det(\mathbf{R}) = 1$$.

  Note: In the following, A1 to An are optional batch dimensions.

  Args:
    matrix: A tensor of shape `[A1, ..., An, K, K]`, where the last two
      dimensions represent a rotation matrix in K-dimensions.
    atol: The absolute tolerance parameter.
    name: A name for this op that defaults to "rotation_matrix_common_is_valid".

  Returns:
    A tensor of type `bool` and shape `[A1, ..., An, 1]` where False indicates
    that the input is not a valid rotation matrix.
  """
  with tf.compat.v1.name_scope(name, "rotation_matrix_common_is_valid",
                               [matrix]):
    matrix = tf.convert_to_tensor(value=matrix)

    shape.check_static(
        tensor=matrix, tensor_name="matrix", has_rank_greater_than=1)
    shape.compare_dimensions(
        tensors=(matrix, matrix),
        tensor_names=("matrix", "matrix"),
        axes=(-1, -2))

    distance_to_unit_determinant = tf.abs(tf.linalg.det(matrix) - 1.)
    # Computes how far the product of the transposed rotation matrix with itself
    # is from the identity matrix.
    ndims = matrix.shape.ndims
    permutation = list(range(ndims - 2)) + [ndims - 1, ndims - 2]
    identity = tf.eye(
        tf.compat.v1.dimension_value(matrix.shape[-1]), dtype=matrix.dtype)
    difference_to_identity = tf.matmul(
        tf.transpose(a=matrix, perm=permutation), matrix) - identity
    norm_diff = tf.norm(tensor=difference_to_identity, axis=(-2, -1))
    # Computes the mask of entries that satisfies all conditions.
    mask = tf.logical_and(distance_to_unit_determinant < atol, norm_diff < atol)
    output = tf.compat.v1.where(
        mask, tf.ones_like(distance_to_unit_determinant, dtype=bool),
        tf.zeros_like(distance_to_unit_determinant, dtype=bool))
    return tf.expand_dims(output, axis=-1)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the precision metric."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import safe_ops
from tensorflow_graphics.util import shape


def _cast_to_int(prediction):
  return tf.cast(x=prediction, dtype=tf.int32)


def evaluate(ground_truth,
             prediction,
             classes=None,
             reduce_average=True,
             prediction_to_category_function=_cast_to_int,
             name=None):
  """Computes the precision metric for the given ground truth and predictions.

  Note:
    In the following, A1 to An are optional batch dimensions, which must be
    broadcast compatible.

  Args:
    ground_truth: A tensor of shape `[A1, ..., An, N]`, where the last axis
      represents the ground truth labels. Will be cast to int32.
    prediction: A tensor of shape `[A1, ..., An, N]`, where the last axis
      represents the predictions (which can be continuous).
    classes: An integer or a list/tuple of integers representing the classes for
      which the precision will be evaluated. In case 'classes' is 'None', the
      number of classes will be inferred from the given labels and the precision
      will be calculated for each of the classes. Defaults to 'None'.
    reduce_average: Whether to calculate the average of the precision for each
      class and return a single precision value. Defaults to true.
    prediction_to_category_function: A function to associate a `prediction` to a
      category. Defaults to rounding down the value of the prediction to the
      nearest integer value.
    name: A name for this op. Defaults to "precision_evaluate".

  Returns:
    A tensor of shape `[A1, ..., An, C]`, where the last axis represents the
    precision calculated for each of the requested classes.

  Raises:
    ValueError: if the shape of `ground_truth`, `prediction` is not supported.
  """
  with tf.compat.v1.name_scope(name, "precision_evaluate",
                               [ground_truth, prediction]):
    ground_truth = tf.cast(
        x=tf.convert_to_tensor(value=ground_truth), dtype=tf.int32)
    prediction = tf.convert_to_tensor(value=prediction)

    shape.compare_batch_dimensions(
        tensors=(ground_truth, prediction),
        tensor_names=("ground_truth", "prediction"),
        last_axes=-1,
        broadcast_compatible=True)

    prediction = prediction_to_category_function(prediction)
    if classes is None:
      num_classes = tf.math.maximum(
          tf.math.reduce_max(input_tensor=ground_truth),
          tf.math.reduce_max(input_tensor=prediction)) + 1
      classes = tf.range(num_classes)
    else:
      classes = tf.convert_to_tensor(value=classes)
      # Make sure classes is a tensor of rank 1.
      classes = tf.reshape(classes, [1]) if tf.rank(classes) == 0 else classes

    # Create a confusion matrix for each of the classes (with dimensions
    # [A1, ..., An, C, N]).
    classes = tf.expand_dims(classes, -1)
    ground_truth_per_class = tf.equal(tf.expand_dims(ground_truth, -2), classes)
    prediction_per_class = tf.equal(tf.expand_dims(prediction, -2), classes)

    # Calculate the precision for each of the classes.
    true_positives = tf.math.reduce_sum(
        input_tensor=tf.cast(
            x=tf.math.logical_and(ground_truth_per_class, prediction_per_class),
            dtype=tf.float32),
        axis=-1)
    total_predicted_positives = tf.math.reduce_sum(
        input_tensor=tf.cast(x=prediction_per_class, dtype=tf.float32), axis=-1)

    precision_per_class = safe_ops.safe_signed_div(true_positives,
                                                   total_predicted_positives)
    if reduce_average:
      return tf.math.reduce_mean(input_tensor=precision_per_class, axis=-1)
    else:
      return precision_per_class


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper routines for mesh unit tests.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np


def create_single_triangle_mesh():
  r"""Creates a single-triangle mesh, in the z=0 plane and facing +z.

  (0,1) 2
        |\
        | \
        |  \
  (0,0) 0---1 (1,0)

  Returns:
    vertices: A [3, 3] float array
    faces: A [1, 3] int array
  """
  vertices = np.array(
      ((0, 0, 0), (1, 0, 0), (0, 1, 0)), dtype=np.float32)
  faces = np.array(((0, 1, 2),), dtype=np.int32)
  return vertices, faces


def create_square_triangle_mesh():
  r"""Creates a square mesh, in the z=0 planse and facing +z.

    # (0,1) 2---3 (1,1)
    #       |\ /|
    #       | 4 |
    #       |/ \|
    # (0,0) 0---1 (1,0)

  Returns:
    vertices: A [5, 3] float array
    faces: A [4, 3] int array
  """
  vertices = np.array(
      ((0, 0, 0), (1, 0, 0), (0, 1, 0), (1, 1, 0), (0.5, 0.5, 0)),
      dtype=np.float32)
  faces = np.array(
      ((0, 1, 4), (1, 3, 4), (3, 2, 4), (2, 0, 4)), dtype=np.int32)
  return vertices, faces



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for shape utility functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import tensorflow as tf

from tensorflow_graphics.util import shape
from tensorflow_graphics.util import test_case


class ShapeTest(test_case.TestCase):

  @parameterized.parameters(
      (None, None, False),
      ((2, 3, 5, 7), (2, 3, 5, 7), True),
      ((1, 3, 5, 7), (2, 3, 5, 7), True),
      ((2, 3, 5, 7), (1, 3, 5, 7), True),
      ((None, 3, 5, 7), (None, 3, 5, 7), True),
      ((2, None, 5, 7), (2, 3, None, 7), True),
      ((1, 3, None, 7), (2, 3, 5, None), True),
      ((None, 3, 5, 7), (None, 3, 5, 7), True),
      ((None, 3, 5, 7), (None, 3, 5, 1), True),
      ((None, 3, 5, 7), (None, 3, 1, 1), True),
      ((None, 3, 5, 7), (None, 1, 1, 1), True),
      ((None, 3, 5, 7), (2, 3, 5, 7), True),
      ((None, 3, 5, 7), (1, 3, 5, 7), True),
      ((None, 3, 5, 7), (None, 5, 7), True),
      ((None, 3, 5, 7), (None, 7), True),
      ((None, 3, 5, 7), (None, 1, 1), True),
      ((None, 3, 5, 7), (None, 1), True),
      ((None, 3, 5, 7), (None,), True),
      ((2, 3, 5, 7), (3, 3, 5, 7), False),
      ((3, 3, 5, 7), (2, 3, 5, 7), False),
      ((None, 3, 5, 7), (None, 3, 5), False),
      ((None, 3, 5, 7), (None, 5), False),
      ((None, 3, 5, 7), (None, 3, 5, 7, 1), False),
      ((None, 3, 5, 7), (None, 2, 5, 7), False),
      ((None, 3, 5, 7), (None, 3, 4, 7), False),
      ((None, 3, 5, 7), (None, 3, 5, 6), False),
  )
  def test_is_broadcast_compatible(self, shape_x, shape_y, broadcastable):
    """Checks if the is_broadcast_compatible function works as expected."""
    if tf.executing_eagerly():
      if (shape_x is None or shape_y is None or None in shape_x or
          None in shape_y):
        return
      shape_x = tf.compat.v1.placeholder_with_default(
          tf.zeros(shape_x, dtype=tf.float32), shape=shape_x).shape
      shape_y = tf.compat.v1.placeholder_with_default(
          tf.zeros(shape_y, dtype=tf.float32), shape=shape_y).shape
    else:
      shape_x = tf.compat.v1.placeholder(shape=shape_x, dtype=tf.float32).shape
      shape_y = tf.compat.v1.placeholder(shape=shape_y, dtype=tf.float32).shape

    self.assertEqual(
        shape.is_broadcast_compatible(shape_x, shape_y), broadcastable)

  @parameterized.parameters(
      (None, None, None),
      ((2, 3, 5, 7), (2, 3, 5, 7), (2, 3, 5, 7)),
      ((1, 3, 5, 7), (2, 3, 5, 7), (2, 3, 5, 7)),
      ((2, 3, 5, 7), (1, 3, 5, 7), (2, 3, 5, 7)),
      ((None, 3, 5, 7), (None, 3, 5, 7), (None, 3, 5, 7)),
      ((2, None, 5, 7), (2, 3, None, 7), (2, 3, 5, 7)),
      ((1, 3, None, 7), (2, 3, 5, None), (2, 3, 5, 7)),
      ((None, 3, 5, 7), (None, 3, 5, 7), (None, 3, 5, 7)),
      ((None, 3, 5, 7), (None, 3, 5, 1), (None, 3, 5, 7)),
      ((None, 3, 5, 7), (None, 3, 1, 1), (None, 3, 5, 7)),
      ((None, 3, 5, 7), (None, 1, 1, 1), (None, 3, 5, 7)),
      ((None, 3, 5, 7), (2, 3, 5, 7), (2, 3, 5, 7)),
      ((None, 3, 5, 7), (1, 3, 5, 7), (None, 3, 5, 7)),
      ((None, 3, 5, 7), (None, 5, 7), (None, 3, 5, 7)),
      ((None, 3, 5, 7), (None, 7), (None, 3, 5, 7)),
      ((None, 3, 5, 7), (None, 1, 1), (None, 3, 5, 7)),
      ((None, 3, 5, 7), (None, 1), (None, 3, 5, 7)),
      ((None, 3, 5, 7), (None,), (None, 3, 5, 7)),
  )
  def test_get_broadcasted_shape(self, shape_x, shape_y, broadcasted_shape):
    """Checks if the get_broadcasted_shape function works as expected."""
    if tf.executing_eagerly():
      if (shape_x is None or shape_y is None or None in shape_x or
          None in shape_y):
        return
      shape_x = tf.compat.v1.placeholder_with_default(
          tf.zeros(shape_x, dtype=tf.float32), shape=shape_x).shape
      shape_y = tf.compat.v1.placeholder_with_default(
          tf.zeros(shape_y, dtype=tf.float32), shape=shape_y).shape
    else:
      shape_x = tf.compat.v1.placeholder(shape=shape_x, dtype=tf.float32).shape
      shape_y = tf.compat.v1.placeholder(shape=shape_y, dtype=tf.float32).shape

    self.assertAllEqual(
        shape.get_broadcasted_shape(shape_x, shape_y), broadcasted_shape)

  @parameterized.parameters(
      ("has_rank must be of type int", "a", None, None, None),
      ("has_rank_greater_than must be of type int", 3, "a", None, None),
      ("has_rank_less_than must be of type int", None, None, "a", None),
      ("must have a rank of 1, but it has", 1, None, None, None),
      ("must have a rank greater than 3, but it has", None, 3, None, None),
      ("must have a rank less than 3, but it has", None, None, 3, None),
      ("has_dim_equals must be of type list or tuple", None, None, None, 0),
      ("has_dim_equals must consist of axis-value pairs", None, None, None,
       (0,)),
      ("has_dim_equals must consist of axis-value pairs", None, None, None,
       ((0, 0, 0),)),
      ("must have exactly 2 dimensions in axis 0", None, None, None, ((0, 2),)),
      ("must have exactly 3 dimensions in axis 1", None, None, None, ((0, 1),
                                                                      (1, 3))),
      ("must have greater than 2 dimensions in axis 1", None, None, None, None,
       ((1, 2),)),
      ("must have greater than 3 dimensions in axis 1", None, None, None, None,
       ((1, 3),)),
      ("must have less than 3 dimensions in axis 2", None, None, None, None,
       None, ((2, 3),)),
      ("must have less than 2 dimensions in axis -1", None, None, None, None,
       None, ((-1, 2),)),
  )
  def test_check_static_raises_exceptions(self,
                                          error_msg,
                                          has_rank,
                                          has_rank_greater_than,
                                          has_rank_less_than,
                                          has_dim_equals,
                                          has_dim_greater_than=None,
                                          has_dim_less_than=None):
    """Tests that check_static raises expected exceptions."""
    self.assert_exception_is_raised(
        shape.check_static,
        error_msg,
        shapes=((1, 2, 3),),
        has_rank=has_rank,
        has_rank_greater_than=has_rank_greater_than,
        has_rank_less_than=has_rank_less_than,
        has_dim_equals=has_dim_equals,
        has_dim_greater_than=has_dim_greater_than,
        has_dim_less_than=has_dim_less_than)

  @parameterized.parameters(
      (((),), 0, None, 1, None),
      (((1,),), 1, 0, 2, (0, 1)),
      (((1, 2),), 2, 1, 3, (0, 1)),
      (((1, 2),), 2, 1, 3, ((0, 1), (1, 2))),
      (((1, 2, 3),), 3, 2, 4, (0, 1)),
      (((1, 2, 3),), 3, 2, 4, ((0, 1), (1, 2), (2, 3))),
  )
  def test_check_static_raises_no_exceptions(self, shapes, has_rank,
                                             has_rank_greater_than,
                                             has_rank_less_than,
                                             has_dim_equals):
    """Tests that check_static works for various inputs."""
    self.assert_exception_is_not_raised(
        shape.check_static,
        shapes=shapes,
        has_rank=has_rank,
        has_rank_greater_than=has_rank_greater_than,
        has_rank_less_than=has_rank_less_than,
        has_dim_equals=has_dim_equals)

  @parameterized.parameters(
      ("tensors must be of type list or tuple", [], (-1), False),
      ("At least 2 tensors are required.", ((1,),), (-1), False),
      ("tensors and last_axes must have the same length", ((1, 1), (1, 1)),
       (0, 0, 0), False),
      ("Some axes are out of bounds.", ((1, 2), (2, 2), (3, 2)),
       (1, 1, 2), False),
      ("Not all batch dimensions are identical.", ((1, 2, 3), (2, 2, 3)),
       (0, 0), False),
      ("Not all batch dimensions are identical.", ((1, 2, 3), (2, 2, 3)),
       (2, 2), False),
      ("Some axes are out of bounds.", ((1, 2), (1, 2), (2, 2)),
       (2, 2, 2), False),
      ("Not all batch dimensions are broadcast-compatible.",
       ((1, 2), (2, 2), (3, 2)), (1, 1, 1), True),
      ("Not all batch dimensions are identical.", ((1, 2, 3), (5, 2, 4),
                                                   (5, 1, 5)), 1, False),
      ("Not all batch dimensions are identical.", ((None, 2, 3), (5, None, 4),
                                                   (5, 1, None)), 1, False),
      ("Not all batch dimensions are identical.",
       ((None, 2, 3), (5, None, 4), (5, 2, None)), (1, 1, 2), False),
      ("Not all batch dimensions are identical.",
       ((1, 2, 3), (1, 2, 3), (1, 2, 3)), (1, 1, 2), False),
      ("Not all batch dimensions are broadcast-compatible.",
       ((None, 2, 3), (5, None, 4), (5, 3, None)), 1, True),
  )
  def test_compare_batch_dimensions_raises_exceptions(self, error_msg,
                                                      tensor_shapes, last_axes,
                                                      broadcast_compatible):
    """Tests that compare_batch_dimensions raises expected exceptions."""
    if not tensor_shapes:
      tensors = 0
    else:
      if all(shape.is_static(tensor_shape) for tensor_shape in tensor_shapes):
        tensors = [tf.ones(tensor_shape) for tensor_shape in tensor_shapes]
      else:
        # Dynamic shapes are not supported in eager mode.
        if tf.executing_eagerly():
          return
        tensors = [
            tf.compat.v1.placeholder(shape=tensor_shape, dtype=tf.float32)
            for tensor_shape in tensor_shapes
        ]
    self.assert_exception_is_raised(
        shape.compare_batch_dimensions,
        error_msg,
        shapes=[],
        tensors=tensors,
        last_axes=last_axes,
        broadcast_compatible=broadcast_compatible)

  @parameterized.parameters(
      (((1, 1), (1, 2)), 0, False, 0),
      (((1, 1), (1, 2)), (0, 0), True, 0),
      (((4, 2, 3), (4, 2, 5, 6)), (1, 1), False, (0, 0)),
      (((4, 2, 3), (1, 2, 5, 6)), 1, True, (0, 0)),
      (((4, 2, 3), (2, 5, 6)), (1, 0), True, 0),
      (((1, 2), (1, 2), (2, 2)), (1, 1, 1), True, 0),
      (((1, 2, 3), (5, 2, 3), (3,)), (2, 2, 0), True, (0, 1, 0)),
      (((1, 2, 3), (5, 2, 4), (5, 1, 5)), 1, True, 0),
      (((None, 2, 3), (5, None, 4), (5, 1, None)), 1, True, 0),
      (((None, 2, 3), (None, 2, 4), (None, 2, None)), 1, False, 0),
      (((None, 2, 3), (None, 2, 4), (None, 2, None)), 1, True, 0),
      (((None, None, 3), (None, None, 4), (None, None, None)), 1, False, 0),
      (((None, None, 3), (None, None, 4), (None, 2, None)), 1, False, 0),
      (((2, None, 3), (2, None, 4), (None, 2, None)), 1, False, 0),
  )
  def test_compare_batch_dimensions_raises_no_exceptions(
      self, tensor_shapes, last_axes, broadcast_compatible, initial_axes):
    """Tests that compare_batch_dimensions works for various inputs."""
    if all(shape.is_static(tensor_shape) for tensor_shape in tensor_shapes):
      tensors = [tf.ones(tensor_shape) for tensor_shape in tensor_shapes]
    else:
      # Dynamic shapes are not supported in eager mode.
      if tf.executing_eagerly():
        return
      tensors = [
          tf.compat.v1.placeholder(shape=tensor_shape, dtype=tf.float32)
          for tensor_shape in tensor_shapes
      ]
    self.assert_exception_is_not_raised(
        shape.compare_batch_dimensions,
        shapes=[],
        tensors=tensors,
        last_axes=last_axes,
        broadcast_compatible=broadcast_compatible,
        initial_axes=initial_axes)

  @parameterized.parameters(
      ("tensors must be of type list or tuple", [], (-1,)),
      ("At least 2 tensors are required.", (1,), (-1,)),
      ("tensors and axes must have the same length", (1, 1), (0, 0, 0)),
      ("Some axes are out of bounds.", ((1, 2), (2, 2), (3, 2)), (1, 1, 2)),
      ("Some axes are out of bounds.", ((1, 2), (2, 2), (3, 2)), (-2, -2, -3)),
      ("must have the same number of dimensions", ((1, 2, 3), (2, 2, 3)),
       (0, 0)),
      ("must have the same number of dimensions", ((1, 2, 5), (2, 2, 3)),
       (1, -1)),
      ("must have the same number of dimensions", ((1, 2), (1, 2), (2, 2)),
       (0, 0, 1)),
      ("must have the same number of dimensions", ((1, 2), (2, 2), (3, 2)),
       (-1, -2, 0)),
  )
  def test_compare_dimensions_raises_exceptions(self, error_msg, tensor_shapes,
                                                axes):
    """Tests that compare_dimensions raises expected exceptions."""
    if not tensor_shapes:
      tensors = 0
    else:
      tensors = [tf.ones(tensor_shape) for tensor_shape in tensor_shapes]
    self.assert_exception_is_raised(
        shape.compare_dimensions,
        error_msg,
        shapes=[],
        tensors=tensors,
        axes=axes)

  @parameterized.parameters(
      ((1, 1), (0, 0)),
      ((1, 1), 0),
      (((4, 2, 3), (4, 2, 5, 6)), (1, 1)),
      (((4, 2, 3), (1, 2, 5, 6)), (1, 1)),
      (((4, 2, 3), (2, 5, 6)), (1, 0)),
      (((1, 2), (1, 2), (2, 2)), (1, 1, 0)),
      (((1, 2, 3), (2, 3), (3,)), (2, 1, 0)),
  )
  def test_compare_dimensions_raises_no_exceptions(self, tensor_shapes, axes):
    """Tests that compare_dimensions works for various inputs."""
    tensors = [tf.ones(tensor_shape) for tensor_shape in tensor_shapes]
    self.assert_exception_is_not_raised(
        shape.compare_dimensions, shapes=[], tensors=tensors, axes=axes)

  @parameterized.parameters(
      ((1,), True),
      ((1, 2), True),
      ((None,), False),
      ((None, 2), False),
  )
  def test_is_static(self, tensor_shape, is_static):
    """Tests that is_static correctly checks if shape is static."""
    if tf.executing_eagerly():
      return
    tensor = tf.compat.v1.placeholder(shape=tensor_shape, dtype=tf.float32)

    with self.subTest(name="tensor_shape_is_list"):
      self.assertEqual(shape.is_static(tensor_shape), is_static)

    with self.subTest(name="tensor_shape"):
      self.assertEqual(shape.is_static(tensor.shape), is_static)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for dual quaternion."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import flagsaver
from absl.testing import parameterized
import tensorflow.compat.v2 as tf

from tensorflow_graphics.geometry.transformation import dual_quaternion
from tensorflow_graphics.geometry.transformation.tests import test_helpers
from tensorflow_graphics.util import test_case


class DualQuaternionTest(test_case.TestCase):

  @parameterized.parameters(
      ((8,),),
      ((None, 8),),
  )
  def test_conjugate_exception_not_raised(self, *shape):
    """Tests that the shape exceptions of conjugate are not raised."""
    self.assert_exception_is_not_raised(dual_quaternion.conjugate, shape)

  @parameterized.parameters(
      ("must have exactly 8 dimensions", (3,)),)
  def test_conjugate_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(dual_quaternion.conjugate, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_conjugate_jacobian_preset(self):
    """Tests the Jacobian of the conjugate function."""
    x_init = test_helpers.generate_preset_test_dual_quaternions()
    self.assert_jacobian_is_correct_fn(dual_quaternion.conjugate, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_conjugate_jacobian_random(self):
    """Tests the Jacobian of the conjugate function."""
    x_init = test_helpers.generate_random_test_dual_quaternions()
    self.assert_jacobian_is_correct_fn(dual_quaternion.conjugate, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_conjugate_preset(self):
    """Tests if the conjugate function is providing correct results."""
    x_init = test_helpers.generate_preset_test_dual_quaternions()
    x = tf.convert_to_tensor(value=x_init)
    y = tf.convert_to_tensor(value=x_init)

    x = dual_quaternion.conjugate(x)
    x_real, x_dual = tf.split(x, (4, 4), axis=-1)

    y_real, y_dual = tf.split(y, (4, 4), axis=-1)
    xyz_y_real, w_y_real = tf.split(y_real, (3, 1), axis=-1)
    xyz_y_dual, w_y_dual = tf.split(y_dual, (3, 1), axis=-1)
    y_real = tf.concat((-xyz_y_real, w_y_real), axis=-1)
    y_dual = tf.concat((-xyz_y_dual, w_y_dual), axis=-1)

    self.assertAllEqual(x_real, y_real)
    self.assertAllEqual(x_dual, y_dual)



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper class for viewing 3D meshes in Colab demos.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from tensorflow_graphics.notebooks import threejs_visualization

SEGMENTATION_COLORMAP = np.array(
    ((165, 242, 12), (89, 12, 89), (165, 89, 165), (242, 242, 165),
     (242, 165, 12), (89, 12, 12), (165, 12, 12), (165, 89, 242), (12, 12, 165),
     (165, 12, 89), (12, 89, 89), (165, 165, 89), (89, 242, 12), (12, 89, 165),
     (242, 242, 89), (165, 165, 165)),
    dtype=np.float32) / 255.0


class Viewer(object):
  """A ThreeJS based viewer class for viewing 3D meshes."""

  def _mesh_from_data(self, data):
    """Creates a dictionary of ThreeJS mesh objects from numpy data."""
    if 'vertices' not in data or 'faces' not in data:
      raise ValueError('Mesh Data must contain vertices and faces')
    vertices = np.asarray(data['vertices'])
    faces = np.asarray(data['faces'])
    material = self.context.THREE.MeshLambertMaterial.new_object({
        'color': 0xfffacd,
        'vertexColors': self.context.THREE.NoColors,
        'side': self.context.THREE.DoubleSide,
    })
    mesh = {'vertices': vertices, 'faces': faces}
    if 'vertex_colors' in data:
      mesh['vertex_colors'] = np.asarray(data['vertex_colors'])
      material = self.context.THREE.MeshLambertMaterial.new_object({
          'color': 0xfffacd,
          'vertexColors': self.context.THREE.VertexColors,
          'side': self.context.THREE.DoubleSide,
      })
    mesh['material'] = material
    return mesh

  def __init__(self, source_mesh_data):
    context = threejs_visualization.build_context()
    self.context = context
    light1 = context.THREE.PointLight.new_object(0x808080)
    light1.position.set(10., 10., 10.)
    light2 = context.THREE.AmbientLight.new_object(0x808080)
    lights = (light1, light2)

    camera = threejs_visualization.build_perspective_camera(
        field_of_view=30, position=(0.0, 0.0, 4.0))

    mesh = self._mesh_from_data(source_mesh_data)
    geometries = threejs_visualization.triangular_mesh_renderer([mesh],
                                                                lights=lights,
                                                                camera=camera,
                                                                width=400,
                                                                height=400)

    self.geometries = geometries



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for asserts."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import test_case


def _pick_random_vector():
  """Creates a random vector with a random shape."""
  tensor_size = np.random.randint(3)
  tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
  return np.random.normal(size=tensor_shape + [4])


class AssertsTest(test_case.TestCase):

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_assert_normalized_exception_not_raised(self, dtype):
    """Checks that assert_normalized raises no exceptions for valid input."""
    vector = _pick_random_vector()
    vector = tf.convert_to_tensor(value=vector, dtype=dtype)
    norm_vector = vector / tf.norm(tensor=vector, axis=-1, keepdims=True)

    self.assert_exception_is_not_raised(
        asserts.assert_normalized, shapes=[], vector=norm_vector)

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_assert_normalized_exception_raised(self, dtype):
    """Checks that assert_normalized raises exceptions for invalid input."""
    vector = _pick_random_vector() + 10.0
    vector = tf.convert_to_tensor(value=vector, dtype=dtype)
    vector = tf.abs(vector)

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(asserts.assert_normalized(vector))

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_assert_normalized_passthrough(self):
    """Checks that the assert is a passthrough when the flag is False."""
    vector_input = _pick_random_vector()

    vector_output = asserts.assert_normalized(vector_input)

    self.assertIs(vector_input, vector_output)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_assert_at_least_k_non_zero_entries_passthrough(self):
    """Checks that the assert is a passthrough when the flag is False."""
    vector_input = _pick_random_vector()

    vector_output = asserts.assert_at_least_k_non_zero_entries(vector_input)

    self.assertIs(vector_input, vector_output)

  @parameterized.parameters(
      (None, None),
      (1e-3, tf.float16),
      (4e-19, tf.float32),
      (4e-154, tf.float64),
  )
  def test_assert_nonzero_norm_exception_not_raised(self, value, dtype):
    """Checks that assert_nonzero_norm works for values above eps."""
    if value is None:
      vector = _pick_random_vector() + 10.0
      vector = tf.convert_to_tensor(value=vector, dtype=dtype)
      vector = tf.abs(vector)
    else:
      vector = tf.constant((value,), dtype=dtype)

    self.assert_exception_is_not_raised(
        asserts.assert_nonzero_norm, shapes=[], vector=vector)

  @parameterized.parameters(
      (1e-4, tf.float16),
      (1e-38, tf.float32),
      (1e-308, tf.float64),
  )
  def test_assert_nonzero_norm_exception_raised(self, value, dtype):
    """Checks that assert_nonzero_norm fails for values below eps."""
    vector = tf.constant((value,), dtype=dtype)

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(asserts.assert_nonzero_norm(vector))

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_assert_nonzero_norm_passthrough(self):
    """Checks that the assert is a passthrough when the flag is False."""
    vector_input = _pick_random_vector()

    vector_output = asserts.assert_nonzero_norm(vector_input)

    self.assertIs(vector_input, vector_output)

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_assert_all_above_exception_not_raised(self, dtype):
    """Checks that assert_all_above raises no exceptions for valid input."""
    vector = _pick_random_vector()
    vector = tf.convert_to_tensor(value=vector, dtype=dtype)

    vector = vector * vector
    vector /= -tf.reduce_max(input_tensor=vector, axis=-1, keepdims=True)
    eps = asserts.select_eps_for_addition(dtype)
    inside_vector = vector + eps
    ones_vector = -tf.ones_like(vector)

    with self.subTest(name="inside_and_open_bounds"):
      self.assert_exception_is_not_raised(
          asserts.assert_all_above,
          shapes=[],
          vector=inside_vector,
          minval=-1.0,
          open_bound=True)

    with self.subTest(name="inside_and_close_bounds"):
      self.assert_exception_is_not_raised(
          asserts.assert_all_above,
          shapes=[],
          vector=inside_vector,
          minval=-1.0,
          open_bound=False)

    with self.subTest(name="exact_and_close_bounds"):
      self.assert_exception_is_not_raised(
          asserts.assert_all_above,
          shapes=[],
          vector=ones_vector,
          minval=-1.0,
          open_bound=False)

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_assert_all_above_exception_raised(self, dtype):
    """Checks that assert_all_above raises exceptions for invalid input."""
    vector = _pick_random_vector()
    vector = tf.convert_to_tensor(value=vector, dtype=dtype)

    vector = vector * vector
    vector /= -tf.reduce_max(input_tensor=vector, axis=-1, keepdims=True)
    eps = asserts.select_eps_for_addition(dtype)
    outside_vector = vector - eps
    ones_vector = -tf.ones_like(vector)

    with self.subTest(name="outside_and_open_bounds"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            asserts.assert_all_above(outside_vector, -1.0, open_bound=True))

    with self.subTest(name="outside_and_close_bounds"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            asserts.assert_all_above(outside_vector, -1.0, open_bound=False))

    with self.subTest(name="exact_and_open_bounds"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            asserts.assert_all_above(ones_vector, -1.0, open_bound=True))

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_assert_all_above_passthrough(self):
    """Checks that the assert is a passthrough when the flag is False."""
    vector_input = _pick_random_vector()

    vector_output = asserts.assert_all_above(vector_input, 1.0)

    self.assertIs(vector_input, vector_output)

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_assert_all_below_exception_not_raised(self, dtype):
    """Checks that assert_all_below raises no exceptions for valid input."""
    vector = _pick_random_vector()
    vector = tf.convert_to_tensor(value=vector, dtype=dtype)

    vector = vector * vector
    vector /= tf.reduce_max(input_tensor=vector, axis=-1, keepdims=True)
    eps = asserts.select_eps_for_addition(dtype)
    inside_vector = vector - eps
    ones_vector = tf.ones_like(vector)

    with self.subTest(name="inside_and_open_bounds"):
      self.assert_exception_is_not_raised(
          asserts.assert_all_below,
          shapes=[],
          vector=inside_vector,
          maxval=1.0,
          open_bound=True)

    with self.subTest(name="inside_and_close_bounds"):
      self.assert_exception_is_not_raised(
          asserts.assert_all_below,
          shapes=[],
          vector=inside_vector,
          maxval=1.0,
          open_bound=False)

    with self.subTest(name="exact_and_close_bounds"):
      self.assert_exception_is_not_raised(
          asserts.assert_all_below,
          shapes=[],
          vector=ones_vector,
          maxval=1.0,
          open_bound=False)

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_assert_all_below_exception_raised(self, dtype):
    """Checks that assert_all_below raises exceptions for invalid input."""
    vector = _pick_random_vector()
    vector = tf.convert_to_tensor(value=vector, dtype=dtype)

    vector = vector * vector
    vector /= tf.reduce_max(input_tensor=vector, axis=-1, keepdims=True)
    eps = asserts.select_eps_for_addition(dtype)
    outside_vector = vector + eps
    ones_vector = tf.ones_like(vector)

    with self.subTest(name="outside_and_open_bounds"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            asserts.assert_all_below(outside_vector, 1.0, open_bound=True))

    with self.subTest(name="outside_and_close_bounds"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            asserts.assert_all_below(outside_vector, 1.0, open_bound=False))

    with self.subTest(name="exact_and_open_bounds"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            asserts.assert_all_below(ones_vector, 1.0, open_bound=True))

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_assert_all_below_passthrough(self):
    """Checks that the assert is a passthrough when the flag is False."""
    vector_input = _pick_random_vector()
    vector_output = asserts.assert_all_below(vector_input, 0.0)

    self.assertIs(vector_input, vector_output)

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_assert_all_in_range_exception_not_raised(self, dtype):
    """Checks that assert_all_in_range raises no exceptions for valid input."""
    vector = _pick_random_vector()
    vector = tf.convert_to_tensor(value=vector, dtype=dtype)

    vector = vector * vector
    vector /= tf.reduce_max(input_tensor=vector, axis=-1, keepdims=True)
    eps = asserts.select_eps_for_addition(dtype)
    inside_vector = vector - eps
    ones_vector = tf.ones_like(vector)

    with self.subTest(name="inside_and_open_bounds"):
      self.assert_exception_is_not_raised(
          asserts.assert_all_in_range,
          shapes=[],
          vector=inside_vector,
          minval=-1.0,
          maxval=1.0,
          open_bounds=True)

    with self.subTest(name="inside_and_close_bounds"):
      self.assert_exception_is_not_raised(
          asserts.assert_all_in_range,
          shapes=[],
          vector=inside_vector,
          minval=-1.0,
          maxval=1.0,
          open_bounds=False)

    with self.subTest(name="exact_and_close_bounds"):
      self.assert_exception_is_not_raised(
          asserts.assert_all_in_range,
          shapes=[],
          vector=ones_vector,
          minval=-1.0,
          maxval=1.0,
          open_bounds=False)

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_assert_all_in_range_exception_raised(self, dtype):
    """Checks that assert_all_in_range raises exceptions for invalid input."""
    vector = _pick_random_vector()
    vector = tf.convert_to_tensor(value=vector, dtype=dtype)

    vector = vector * vector
    vector /= tf.reduce_max(input_tensor=vector, axis=-1, keepdims=True)
    eps = asserts.select_eps_for_addition(dtype)
    outside_vector = vector + eps
    ones_vector = tf.ones_like(vector)

    with self.subTest(name="outside_and_open_bounds"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            asserts.assert_all_in_range(
                outside_vector, -1.0, 1.0, open_bounds=True))

    with self.subTest(name="outside_and_close_bounds"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            asserts.assert_all_in_range(
                outside_vector, -1.0, 1.0, open_bounds=False))

    with self.subTest(name="exact_and_open_bounds"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(
            asserts.assert_all_in_range(
                ones_vector, -1.0, 1.0, open_bounds=True))

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_assert_all_in_range_passthrough(self):
    """Checks that the assert is a passthrough when the flag is False."""
    vector_input = _pick_random_vector()

    vector_output = asserts.assert_all_in_range(vector_input, -1.0, 1.0)

    self.assertIs(vector_input, vector_output)

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_select_eps_for_division(self, dtype):
    """Checks that select_eps_for_division does not cause Inf values."""
    a = tf.constant(1.0, dtype=dtype)
    eps = asserts.select_eps_for_division(dtype)

    self.assert_exception_is_not_raised(
        asserts.assert_no_infs_or_nans, shapes=[], tensor=a / eps)

  @parameterized.parameters(tf.float16, tf.float32, tf.float64)
  def test_select_eps_for_addition(self, dtype):
    """Checks that select_eps_for_addition returns large enough eps."""
    a = tf.constant(1.0, dtype=dtype)
    eps = asserts.select_eps_for_addition(dtype)

    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(tf.compat.v1.assert_equal(a, a + eps))

  @parameterized.parameters((np.NaN,), (np.inf,))
  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_assert_no_infs_or_nans_passthrough(self, value):
    """Checks that the assert is a passthrough when the flag is False."""
    vector_input = (value,)

    vector_output = asserts.assert_no_infs_or_nans(vector_input)

    self.assertIs(vector_input, vector_output)

  @parameterized.parameters((np.NaN,), (np.inf,))
  def test_assert_no_infs_or_nans_raises_exception_for_nan(self, value):
    """Checks that the assert works for `Inf` or `NaN` values."""
    vector_input = (value,)

    with self.assertRaisesRegex(  # pylint: disable=g-error-prone-assert-raises
        tf.errors.InvalidArgumentError, "Inf or NaN detected."):
      self.evaluate(asserts.assert_no_infs_or_nans(vector_input))

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_assert_binary_passthrough(self):
    """Checks that the assert is a passthrough when the flag is False."""
    vector_input = _pick_random_vector()

    vector_output = asserts.assert_binary(vector_input)

    self.assertIs(vector_input, vector_output)

  # pylint: disable=g-error-prone-assert-raises
  @parameterized.parameters(tf.float16, tf.float32, tf.float64, tf.int16,
                            tf.int32, tf.int64)
  def test_assert_binary_exception_raised(self, dtype):
    """Checks that assert_binary raises exceptions for invalid input."""
    tensor_size = np.random.randint(3) + 1
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    num_elements = np.prod(tensor_shape)

    # Vector with all ones except for a single negative entry.
    vector_with_negative = np.ones(num_elements)
    vector_with_negative[np.random.randint(num_elements)] = -1
    vector_with_negative = vector_with_negative.reshape(tensor_shape)
    vector_with_negative = tf.convert_to_tensor(
        value=vector_with_negative, dtype=dtype)

    # Vector with all zeros except for a single 0.5 (or 2 in case dtype=int).
    vector = np.zeros(num_elements)
    vector[np.random.randint(num_elements)] = 2
    vector = vector.reshape(tensor_shape)
    vector = tf.convert_to_tensor(value=vector, dtype=dtype)
    vector = vector - tf.compat.v1.div(vector, 4) * 3

    with self.subTest(name="has_negative_number"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(asserts.assert_binary(vector_with_negative))

    with self.subTest(name="has_non_binary_number"):
      with self.assertRaises(tf.errors.InvalidArgumentError):
        self.evaluate(asserts.assert_binary(vector))

  @parameterized.parameters(tf.float16, tf.float32, tf.float64, tf.int16,
                            tf.int32, tf.int64)
  def test_assert_binary_exception_not_raised(self, dtype):
    """Checks that assert_binary raises no exceptions for valid input."""
    tensor_size = np.random.randint(3) + 1
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()

    # Vector with random zeros and ones.
    vector = np.random.randint(2, size=tensor_shape)
    vector = tf.convert_to_tensor(value=vector, dtype=dtype)

    self.assert_exception_is_not_raised(
        asserts.assert_binary, shapes=[], tensor=vector)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for normals."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.representation.mesh import normals
from tensorflow_graphics.util import test_case


class MeshTest(test_case.TestCase):

  @parameterized.parameters(
      (((None, 3), (None, 3)), (tf.float32, tf.int32)),
      (((3, 6, 3), (3, 5, 4)), (tf.float32, tf.int32)),
  )
  def test_gather_faces_exception_not_raised(self, shapes, dtypes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(normals.gather_faces, shapes, dtypes)

  @parameterized.parameters(
      ("Not all batch dimensions are identical", (3, 5, 4, 4), (1, 2, 4, 4)),
      ("Not all batch dimensions are identical", (5, 4, 4), (1, 2, 4, 4)),
      ("Not all batch dimensions are identical", (3, 5, 4, 4), (2, 4, 4)),
      ("vertices must have a rank greater than 1", (4,), (1, 2, 4, 4)),
      ("indices must have a rank greater than 1", (3, 5, 4, 4), (4,)),
  )
  def test_gather_faces_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(normals.gather_faces, error_msg, shapes)

  def test_gather_faces_jacobian_random(self):
    """Test the Jacobian of the face extraction function."""
    tensor_size = np.random.randint(2, 5)
    tensor_shape = np.random.randint(1, 5, size=tensor_size).tolist()
    vertex_init = np.random.random(size=tensor_shape)
    indices_init = np.random.randint(0, tensor_shape[-2], size=tensor_shape)
    indices_tensor = tf.convert_to_tensor(value=indices_init)

    def gather_faces(vertex_tensor):
      return normals.gather_faces(vertex_tensor, indices_tensor)

    self.assert_jacobian_is_correct_fn(gather_faces, [vertex_init])

  @parameterized.parameters(
      ((((0.,), (1.,)), ((1, 0),)), ((((1.,), (0.,)),),)),
      ((((0., 1.), (2., 3.)), ((1, 0),)), ((((2., 3.), (0., 1.)),),)),
      ((((0., 1., 2.), (3., 4., 5.)), ((1, 0),)), ((((3., 4., 5.),
                                                     (0., 1., 2.)),),)),
  )
  def test_gather_faces_preset(self, test_inputs, test_outputs):
    """Tests the extraction of mesh faces."""
    self.assert_output_is_correct(
        normals.gather_faces, test_inputs, test_outputs, tile=False)

  def test_gather_faces_random(self):
    """Tests the extraction of mesh faces."""
    tensor_size = np.random.randint(3, 5)
    tensor_shape = np.random.randint(1, 5, size=tensor_size).tolist()
    vertices = np.random.random(size=tensor_shape)
    indices = np.arange(tensor_shape[-2])
    indices = indices.reshape([1] * (tensor_size - 1) + [-1])
    indices = np.tile(indices, tensor_shape[:-2] + [1, 1])
    expected = np.expand_dims(vertices, -3)

    self.assertAllClose(
        normals.gather_faces(vertices, indices), expected, rtol=1e-3)

  @parameterized.parameters(
      (((None, 4, 3),), (tf.float32,)),
      (((4, 3),), (tf.float32,)),
      (((3, 4, 3),), (tf.float32,)),
  )
  def test_face_normals_exception_not_raised(self, shapes, dtypes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(normals.face_normals, shapes, dtypes)

  @parameterized.parameters(
      ("faces must have a rank greater than 1.", (3,)),
      ("faces must have greater than 2 dimensions in axis -2", (2, 3)),
      ("faces must have exactly 3 dimensions in axis -1.", (5, 2)),
  )
  def test_face_normals_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(normals.face_normals, error_msg, shapes)

  def test_face_normals_jacobian_random(self):
    """Test the Jacobian of the face normals function."""
    tensor_vertex_size = np.random.randint(1, 3)
    tensor_out_shape = np.random.randint(1, 5, size=tensor_vertex_size)
    tensor_out_shape = tensor_out_shape.tolist()
    tensor_vertex_shape = list(tensor_out_shape)
    tensor_vertex_shape[-1] *= 3
    tensor_index_shape = tensor_out_shape[-1]
    vertex_init = np.random.random(size=tensor_vertex_shape + [3])
    index_init = np.arange(tensor_vertex_shape[-1])
    np.random.shuffle(index_init)
    index_init = np.reshape(index_init, newshape=[1] * \
                            (tensor_vertex_size - 1) + \
                            [tensor_index_shape, 3])
    index_init = np.tile(index_init, tensor_vertex_shape[:-1] + [1, 1])
    index_tensor = tf.convert_to_tensor(value=index_init)

    def face_normals(vertex_tensor):
      face_tensor = normals.gather_faces(vertex_tensor, index_tensor)
      return normals.face_normals(face_tensor)

    self.assert_jacobian_is_correct_fn(
        face_normals, [vertex_init], atol=1e-4, delta=1e-9)

  @parameterized.parameters(
      ((((0., 0., 0.), (1., 0., 0.), (0., 1., 0.)), ((0, 1, 2),)),
       (((0., 0., 1.),),)),
      ((((0., 0., 0.), (0., 0., 1.), (1., 0., 0.)), ((0, 1, 2),)),
       (((0., 1., 0.),),)),
      ((((0., 0., 0.), (0., 1., 0.), (0., 0., 1.)), ((0, 1, 2),)),
       (((1., 0., 0.),),)),
      ((((0., -2., -2.), (0, -2., 2.), (0., 2., 2.), (0., 2., -2.)),
        ((0, 1, 2, 3),)), (((-1., 0., 0.),),)),
  )
  def test_face_normals_preset(self, test_inputs, test_outputs):
    """Tests the computation of mesh face normals."""
    faces = normals.gather_faces(*test_inputs[:2])
    test_inputs = [faces] + list(test_inputs[2:])

    self.assert_output_is_correct(
        normals.face_normals, test_inputs, test_outputs, tile=False)

  def test_face_normals_random(self):
    """Tests the computation of mesh face normals in each axis."""
    tensor_vertex_size = np.random.randint(1, 3)
    tensor_out_shape = np.random.randint(1, 5, size=tensor_vertex_size)
    tensor_out_shape = tensor_out_shape.tolist()
    tensor_vertex_shape = list(tensor_out_shape)
    tensor_vertex_shape[-1] *= 3
    tensor_index_shape = tensor_out_shape[-1]

    for i in range(3):
      vertices = np.random.random(size=tensor_vertex_shape + [3])
      indices = np.arange(tensor_vertex_shape[-1])
      np.random.shuffle(indices)
      indices = np.reshape(indices,
                           newshape=[1] * (tensor_vertex_size - 1) \
                           + [tensor_index_shape, 3])
      indices = np.tile(indices, tensor_vertex_shape[:-1] + [1, 1])
      vertices[..., i] = 0.
      expected = np.zeros(shape=tensor_out_shape + [3], dtype=vertices.dtype)
      expected[..., i] = 1.
      faces = normals.gather_faces(vertices, indices)

      self.assertAllClose(
          tf.abs(normals.face_normals(faces)), expected, rtol=1e-3)

  @parameterized.parameters(
      (((4, 3), (5, 3)), (tf.float32, tf.int32)),
      (((None, 3), (None, 3)), (tf.float32, tf.int32)),
      (((3, None, 3), (3, None, 5)), (tf.float32, tf.int32)),
      (((3, 6, 3), (3, 5, 5)), (tf.float32, tf.int32)),
  )
  def test_vertex_normals_exception_not_raised(self, shapes, dtypes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(normals.vertex_normals, shapes, dtypes)

  @parameterized.parameters(
      ("Not all batch dimensions are broadcast-compatible.", (3, 5, 4, 3),
       (1, 2, 4, 3)),
      ("Not all batch dimensions are broadcast-compatible.", (2, 200, 3),
       (4, 100, 3)),
      ("Not all batch dimensions are broadcast-compatible.", (5, 4, 3),
       (1, 2, 4, 3)),
      ("Not all batch dimensions are broadcast-compatible.", (3, 5, 4, 3),
       (2, 4, 3)),
      ("vertices must have a rank greater than 1.", (3,), (1, 2, 4, 3)),
      ("indices must have a rank greater than 1.", (3, 5, 4, 3), (3,)),
      ("vertices must have exactly 3 dimensions in axis -1.", (3, 5, 4, 2),
       (3, 5, 4, 3)),
      ("indices must have greater than 2 dimensions in axis -1.", (3, 5, 4, 3),
       (3, 5, 4, 2)),
      ("'indices' must have specified batch dimensions.", (None, 6, 3),
       (None, 5, 5)),
  )
  def test_vertex_normals_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(normals.vertex_normals, error_msg, shapes)

  def test_vertex_normals_jacobian_random(self):
    """Test the Jacobian of the vertex normals function."""
    tensor_vertex_size = np.random.randint(1, 3)
    tensor_out_shape = np.random.randint(1, 5, size=tensor_vertex_size)
    tensor_out_shape = tensor_out_shape.tolist()
    vertex_axis = np.array(((0., 0., 1), (1., 0., 0.), (0., 1., 0.),
                            (0., 0., -1.), (-1., 0., 0.), (0., -1., 0.)),
                           dtype=np.float32)
    vertex_axis = vertex_axis.reshape([1] * tensor_vertex_size + [6, 3])
    faces = np.array(((0, 1, 2), (0, 2, 4), (0, 4, 5), (0, 5, 1), (3, 2, 1),
                      (3, 4, 2), (3, 5, 4), (3, 1, 5)),
                     dtype=np.int32)
    faces = faces.reshape([1] * tensor_vertex_size + [8, 3])
    index_init = np.tile(faces, tensor_out_shape + [1, 1])
    vertex_scale = np.random.uniform(0.5, 5., tensor_out_shape + [1] * 2)
    vertex_init = vertex_axis * vertex_scale
    index_tensor = tf.convert_to_tensor(value=index_init)

    def vertex_normals(vertex_tensor):
      return normals.vertex_normals(vertex_tensor, index_tensor)

    self.assert_jacobian_is_correct_fn(vertex_normals, [vertex_init])

  @parameterized.parameters(
      (((((-1., -1., 1.), (-1., 1., 1.), (-1., -1., -1.), (-1., 1., -1.),
          (1., -1., 1.), (1., 1., 1.), (1., -1., -1.), (1., 1., -1.)),),
        (((1, 2, 0), (3, 6, 2), (7, 4, 6), (5, 0, 4), (6, 0, 2), (3, 5, 7),
          (1, 3, 2), (3, 7, 6), (7, 5, 4), (5, 1, 0), (6, 4, 0), (3, 1, 5)),)),
       ((((-0.3333333134651184, -0.6666666269302368, 0.6666666269302368),
          (-0.8164965510368347, 0.40824827551841736, 0.40824827551841736),
          (-0.8164965510368347, -0.40824827551841736, -0.40824827551841736),
          (-0.3333333134651184, 0.6666666269302368, -0.6666666269302368),
          (0.8164965510368347, -0.40824827551841736, 0.40824827551841736),
          (0.3333333134651184, 0.6666666269302368, 0.6666666269302368),
          (0.3333333134651184, -0.6666666269302368, -0.6666666269302368),
          (0.8164965510368347, 0.40824827551841736, -0.40824827551841736)),),)),
  )
  def test_vertex_normals_preset(self, test_inputs, test_outputs):
    """Tests the computation of vertex normals."""
    self.assert_output_is_correct(
        normals.vertex_normals, test_inputs, test_outputs, tile=False)

  def test_vertex_normals_random(self):
    """Tests the computation of vertex normals for a regular octahedral."""
    tensor_vertex_size = np.random.randint(1, 3)
    tensor_out_shape = np.random.randint(1, 5, size=tensor_vertex_size)
    tensor_out_shape = tensor_out_shape.tolist()

    with self.subTest(name="triangular_faces"):
      vertex_on_axes = np.array(((0., 0., 1), (1., 0., 0.), (0., 1., 0.),
                                 (0., 0., -1.), (-1., 0., 0.), (0., -1., 0.)),
                                dtype=np.float32)
      vertex_on_axes = vertex_on_axes.reshape([1] * tensor_vertex_size + [6, 3])
      index_init = np.array(((0, 1, 2), (0, 2, 4), (0, 4, 5), (0, 5, 1),
                             (3, 2, 1), (3, 4, 2), (3, 5, 4), (3, 1, 5)),
                            dtype=np.int32)
      index_init = index_init.reshape([1] * tensor_vertex_size + [8, 3])
      index_init = np.tile(index_init, tensor_out_shape + [1, 1])
      vertex_scale = np.random.uniform(0.5, 5., tensor_out_shape + [1] * 2)
      vertex_init = vertex_on_axes * vertex_scale
      expected = vertex_on_axes * (vertex_scale * 0. + 1.)

      vertex_tensor = tf.convert_to_tensor(value=vertex_init)
      index_tensor = tf.convert_to_tensor(value=index_init)

      self.assertAllClose(
          normals.vertex_normals(vertex_tensor, index_tensor), expected)

    with self.subTest(name="polygon_faces"):
      num_vertices = np.random.randint(4, 8)
      poly_vertices = []
      rad_step = np.pi * 2. / num_vertices
      for i in range(num_vertices):
        poly_vertices.append([np.cos(i * rad_step), np.sin(i * rad_step), 0])
      vertex_init = np.array(poly_vertices, dtype=np.float32)
      vertex_init = vertex_init.reshape([1] * tensor_vertex_size + [-1, 3])
      vertex_init = vertex_init * vertex_scale
      index_init = np.arange(num_vertices, dtype=np.int32)
      index_init = index_init.reshape([1] * tensor_vertex_size + [1, -1])
      index_init = np.tile(index_init, tensor_out_shape + [1, 1])
      expected = np.array((0., 0., 1.), dtype=np.float32)
      expected = expected.reshape([1] * tensor_vertex_size + [1, 3])
      expected = np.tile(expected, tensor_out_shape + [num_vertices, 1])
      vertex_tensor = tf.convert_to_tensor(value=vertex_init)
      index_tensor = tf.convert_to_tensor(value=index_init)

      self.assertAllClose(
          normals.vertex_normals(vertex_tensor, index_tensor), expected)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for perspective camera functionalities."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import sys

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.rendering.camera import perspective
from tensorflow_graphics.util import test_case


class PerspectiveTest(test_case.TestCase):

  @parameterized.parameters(
      ("must have exactly 4 dimensions in axis -1", (4, 3)),
      ("must have exactly 4 dimensions in axis -2", (5, 4)),
      ("must have exactly 4 dimensions in axis -2", (None, 4)),
      ("must have exactly 4 dimensions in axis -1", (4, None)),
  )
  def test_parameters_from_right_handed_shape_exception_raised(
      self, error_msg, *shapes):
    """Checks the inputs of the from_right_handed_shape function."""
    self.assert_exception_is_raised(perspective.parameters_from_right_handed,
                                    error_msg, shapes)

  @parameterized.parameters(
      ((4, 4),),
      ((None, 4, 4),),
      ((None, None, 4, 4),),
  )
  def test_parameters_from_right_handed_shape_exception_not_raised(
      self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        perspective.parameters_from_right_handed, shapes)

  def test_parameters_from_right_handed_random(self):
    """Tests that parameters_from_right_handed returns the expected values."""
    tensor_size = np.random.randint(2, 4)
    tensor_shape = np.random.randint(2, 5, size=(tensor_size)).tolist()
    vertical_field_of_view_gt = np.random.uniform(
        sys.float_info.epsilon, np.pi - sys.float_info.epsilon,
        tensor_shape + [1])
    aspect_ratio_gt = np.random.uniform(0.1, 10.0, tensor_shape + [1])
    near_gt = np.random.uniform(0.1, 100.0, tensor_shape + [1])
    far_gt = near_gt + np.random.uniform(0.1, 100.0, tensor_shape + [1])
    projection_matrix = perspective.right_handed(vertical_field_of_view_gt,
                                                 aspect_ratio_gt, near_gt,
                                                 far_gt)

    vertical_field_of_view_pred, aspect_ratio_pred, near_pred, far_pred = perspective.parameters_from_right_handed(
        projection_matrix)

    with self.subTest(name="vertical_field_of_view"):
      self.assertAllClose(vertical_field_of_view_gt,
                          vertical_field_of_view_pred)

    with self.subTest(name="aspect_ratio"):
      self.assertAllClose(aspect_ratio_gt, aspect_ratio_pred)

    with self.subTest(name="near_plane"):
      self.assertAllClose(near_gt, near_pred)

    with self.subTest(name="far_plane"):
      self.assertAllClose(far_gt, far_pred)

  def test_parameters_from_right_handed_jacobian_random(self):
    """Tests the Jacobian of parameters_from_right_handed."""
    tensor_size = np.random.randint(2, 4)
    tensor_shape = np.random.randint(2, 5, size=(tensor_size)).tolist()
    vertical_field_of_view = np.random.uniform(sys.float_info.epsilon,
                                               np.pi - sys.float_info.epsilon,
                                               tensor_shape + [1])
    aspect_ratio = np.random.uniform(0.1, 10.0, tensor_shape + [1])
    near = np.random.uniform(0.1, 100.0, tensor_shape + [1])
    far = near + np.random.uniform(0.1, 100.0, tensor_shape + [1])
    projection_matrix = perspective.right_handed(vertical_field_of_view,
                                                 aspect_ratio, near, far)

    with self.subTest(name="vertical_field_of_view"):
      self.assert_jacobian_is_finite_fn(
          lambda x: perspective.parameters_from_right_handed(x)[0],
          [projection_matrix])

    with self.subTest(name="aspect_ratio"):
      self.assert_jacobian_is_finite_fn(
          lambda x: perspective.parameters_from_right_handed(x)[1],
          [projection_matrix])

    with self.subTest(name="near_plane"):
      self.assert_jacobian_is_finite_fn(
          lambda x: perspective.parameters_from_right_handed(x)[2],
          [projection_matrix])

    with self.subTest(name="far_plane"):
      self.assert_jacobian_is_finite_fn(
          lambda x: perspective.parameters_from_right_handed(x)[3],
          [projection_matrix])

  def test_perspective_right_handed_preset(self):
    """Tests that perspective_right_handed generates expected results."""
    vertical_field_of_view = ((60.0 * math.pi / 180.0,),
                              (50.0 * math.pi / 180.0,))
    aspect_ratio = ((1.5,), (1.1,))
    near = ((1.0,), (1.2,))
    far = ((10.0,), (5.0,))

    pred = perspective.right_handed(vertical_field_of_view, aspect_ratio, near,
                                    far)
    gt = (((1.15470052, 0.0, 0.0, 0.0), (0.0, 1.73205066, 0.0, 0.0),
           (0.0, 0.0, -1.22222221, -2.22222233), (0.0, 0.0, -1.0, 0.0)),
          ((1.9495517, 0.0, 0.0, 0.0), (0.0, 2.14450693, 0.0, 0.0),
           (0.0, 0.0, -1.63157892, -3.15789485), (0.0, 0.0, -1.0, 0.0)))
    self.assertAllClose(pred, gt)

  @parameterized.parameters(
      ((1,), (1,), (1,), (1,)),
      ((None, 1), (None, 1), (None, 1), (None, 1)),
      ((None, 3, 1), (None, 3, 1), (None, 3, 1), (None, 3, 1)),
  )
  def test_perspective_right_handed_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(perspective.right_handed, shapes)

  @parameterized.parameters(
      ("Not all batch dimensions are identical", (1,), (3, 1), (3, 1), (3, 1)),
      ("Not all batch dimensions are identical", (3, 1), (None, 3, 1), (3, 1),
       (3, 1)),
  )
  def test_perspective_right_handed_shape_exception_raised(
      self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(perspective.right_handed, error_msg, shapes)

  @parameterized.parameters(
      ((1.0,),
       (1.0,), np.random.uniform(-1.0, 0.0, size=(1,)).astype(np.float32),
       (1.0,)),
      ((1.0,), (1.0,), (0.0,), (1.0,)),
      ((1.0,), np.random.uniform(-1.0, 0.0, size=(1,)).astype(np.float32),
       (0.1,), (1.0,)),
      ((1.0,), (0.0,), (0.1,), (1.0,)),
      ((1.0,),
       (1.0,), np.random.uniform(1.0, 2.0, size=(1,)).astype(np.float32),
       np.random.uniform(0.1, 0.5, size=(1,)).astype(np.float32)),
      ((1.0,), (1.0,), (0.1,), (0.1,)),
      (np.random.uniform(-math.pi, 0.0, size=(1,)).astype(np.float32), (1.0,),
       (0.1,), (1.0,)),
      (np.random.uniform(math.pi, 2.0 * math.pi, size=(1,)).astype(np.float32),
       (1.0,), (0.1,), (1.0,)),
      ((0.0,), (1.0,), (0.1,), (1.0,)),
      ((math.pi,), (1.0,), (0.1,), (1.0,)),
  )
  def test_perspective_right_handed_valid_range_exception_raised(
      self, vertical_field_of_view, aspect_ratio, near, far):
    """Tests that an exception is raised with out of bounds values."""
    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          perspective.right_handed(vertical_field_of_view, aspect_ratio, near,
                                   far))

  def test_perspective_right_handed_cross_jacobian_preset(self):
    """Tests the Jacobian of perspective_right_handed."""
    vertical_field_of_view_init = np.array((1.0,))
    aspect_ratio_init = np.array((1.0,))
    near_init = np.array((1.0,))
    far_init = np.array((10.0,))

    self.assert_jacobian_is_correct_fn(
        perspective.right_handed,
        [vertical_field_of_view_init, aspect_ratio_init, near_init, far_init])

  def test_perspective_right_handed_cross_jacobian_random(self):
    """Tests the Jacobian of perspective_right_handed."""
    tensor_size = np.random.randint(1, 3)
    tensor_shape = np.random.randint(1, 5, size=(tensor_size)).tolist()
    eps = np.finfo(np.float64).eps
    vertical_field_of_view_init = np.random.uniform(
        eps, math.pi - eps, size=tensor_shape + [1])
    aspect_ratio_init = np.random.uniform(eps, 100.0, size=tensor_shape + [1])
    near_init = np.random.uniform(eps, 10.0, size=tensor_shape + [1])
    far_init = np.random.uniform(10 + eps, 100.0, size=tensor_shape + [1])

    self.assert_jacobian_is_correct_fn(
        perspective.right_handed,
        [vertical_field_of_view_init, aspect_ratio_init, near_init, far_init])

  @parameterized.parameters(
      ((3, 3),),
      ((3, 3, 3),),
      ((None, 3, 3),),
  )
  def test_intrinsics_from_matrix_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(perspective.intrinsics_from_matrix,
                                        shapes)

  @parameterized.parameters(
      ("must have a rank greater than 1", (3,)),
      ("must have exactly 3 dimensions in axis -2", (None, 3)),
      ("must have exactly 3 dimensions in axis -1", (3, None)),
  )
  def test_intrinsics_from_matrix_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(perspective.intrinsics_from_matrix,
                                    error_msg, shapes)

  @parameterized.parameters(
      ((((0., 0., 0.), (0., 0., 0.), (0., 0., 1.)),), ((0., 0.), (0., 0.))),
      ((((1., 0., 3.), (0., 2., 4.), (0., 0., 1.)),), ((1., 2.), (3., 4.))),
  )
  def test_intrinsics_from_matrix_preset(self, test_inputs, test_outputs):
    """Tests that intrinsics_from_matrix gives the correct result."""
    self.assert_output_is_correct(perspective.intrinsics_from_matrix,
                                  test_inputs, test_outputs)

  def test_intrinsics_from_matrix_to_intrinsics_random(self):
    """Tests that converting intrinsics to a matrix and back is consistent."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    random_focal = np.random.normal(size=tensor_shape + [2])
    random_principal_point = np.random.normal(size=tensor_shape + [2])

    matrix = perspective.matrix_from_intrinsics(random_focal,
                                                random_principal_point)
    focal, principal_point = perspective.intrinsics_from_matrix(matrix)

    self.assertAllClose(random_focal, focal, rtol=1e-3)
    self.assertAllClose(random_principal_point, principal_point, rtol=1e-3)

  @parameterized.parameters(
      ((2,), (2,)),
      ((2, 2), (2, 2)),
      ((None, 2), (None, 2)),
  )
  def test_matrix_from_intrinsics_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(perspective.matrix_from_intrinsics,
                                        shapes)

  @parameterized.parameters(
      ("must have exactly 2 dimensions in axis -1", (None,), (2,)),
      ("must have exactly 2 dimensions in axis -1", (2,), (None,)),
      ("Not all batch dimensions are identical.", (3, 2), (2, 2)),
  )
  def test_matrix_from_intrinsics_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(perspective.matrix_from_intrinsics,
                                    error_msg, shapes)

  @parameterized.parameters(
      (((0., 0.), (0., 0.)), (((0., 0., 0.), (0., 0., 0.), (0., 0., 1.)),)),
      (((1., 2.), (3., 4.)), (((1., 0., 3.), (0., 2., 4.), (0., 0., 1.)),)),
  )
  def test_matrix_from_intrinsics_preset(self, test_inputs, test_outputs):
    """Tests that matrix_from_intrinsics gives the correct result."""
    self.assert_output_is_correct(perspective.matrix_from_intrinsics,
                                  test_inputs, test_outputs)

  def test_matrix_from_intrinsics_to_matrix_random(self):
    """Tests that converting a matrix to intrinsics and back is consistent."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    random_focal = np.random.normal(size=tensor_shape + [2])
    random_principal_point = np.random.normal(size=tensor_shape + [2])
    fx = random_focal[..., 0]
    fy = random_focal[..., 1]
    cx = random_principal_point[..., 0]
    cy = random_principal_point[..., 1]
    zero = np.zeros_like(fx)
    one = np.ones_like(fx)
    random_matrix = np.stack((fx, zero, cx, zero, fy, cy, zero, zero, one),
                             axis=-1).reshape(tensor_shape + [3, 3])

    focal, principal_point = perspective.intrinsics_from_matrix(random_matrix)
    matrix = perspective.matrix_from_intrinsics(focal, principal_point)

    self.assertAllClose(random_matrix, matrix, rtol=1e-3)

  @parameterized.parameters(
      ((3,), (2,), (2,)),
      ((2, 3), (2, 2), (2, 2)),
      ((2, 3), (2,), (2,)),
      ((None, 3), (None, 2), (None, 2)),
  )
  def test_project_exception_not_exception_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(perspective.project, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (None,), (2,), (2,)),
      ("must have exactly 2 dimensions in axis -1", (3,), (None,), (2,)),
      ("must have exactly 2 dimensions in axis -1", (3,), (2,), (None,)),
      ("Not all batch dimensions are broadcast-compatible.", (3, 3), (2, 2),
       (2, 2)),
  )
  def test_project_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(perspective.project, error_msg, shape)

  @parameterized.parameters(
      (((0., 0., 1.), (1., 1.), (0., 0.)), ((0., 0.),)),
      (((4., 2., 1.), (1., 1.), (-4., -2.)), ((0., 0.),)),
      (((4., 2., 10.), (1., 1.), (-.4, -.2)), ((0., 0.),)),
      (((4., 2., 10.), (2., 1.), (-.8, -.2)), ((0., 0.),)),
      (((4., 2., 10.), (2., 1.), (-.8, 0.)), ((0., .2),)),
  )
  def test_project_preset(self, test_inputs, test_outputs):
    """Tests that the project function gives the correct result."""
    self.assert_output_is_correct(perspective.project, test_inputs,
                                  test_outputs)

  def test_project_unproject_random(self):
    """Tests that projecting and unprojecting gives an identity mapping."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    random_point_3d = np.random.normal(size=tensor_shape + [3])
    random_focal = np.random.normal(size=tensor_shape + [2])
    random_principal_point = np.random.normal(size=tensor_shape + [2])
    random_depth = np.expand_dims(random_point_3d[..., 2], axis=-1)

    point_2d = perspective.project(random_point_3d, random_focal,
                                   random_principal_point)
    point_3d = perspective.unproject(point_2d, random_depth, random_focal,
                                     random_principal_point)

    self.assertAllClose(random_point_3d, point_3d, rtol=1e-3)

  def test_project_ray_random(self):
    """Tests that that ray is pointing toward the correct location."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    random_point_3d = np.random.normal(size=tensor_shape + [3])
    random_focal = np.random.normal(size=tensor_shape + [2])
    random_principal_point = np.random.normal(size=tensor_shape + [2])
    random_depth = np.expand_dims(random_point_3d[..., 2], axis=-1)

    point_2d = perspective.project(random_point_3d, random_focal,
                                   random_principal_point)
    ray_3d = perspective.ray(point_2d, random_focal, random_principal_point)
    ray_3d = random_depth * ray_3d

    self.assertAllClose(random_point_3d, ray_3d, rtol=1e-3)

  @parameterized.parameters(
      ((2,), (2,), (2,)),
      ((2, 2), (2, 2), (2, 2)),
      ((3, 2), (1, 2), (2,)),
      ((None, 2), (None, 2), (None, 2)),
  )
  def test_ray_exception_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(perspective.ray, shapes)

  @parameterized.parameters(
      ("must have exactly 2 dimensions in axis -1", (None,), (2,), (2,)),
      ("must have exactly 2 dimensions in axis -1", (2,), (None,), (2,)),
      ("must have exactly 2 dimensions in axis -1", (2,), (2,), (None,)),
      ("Not all batch dimensions are broadcast-compatible.", (3, 2), (1, 2),
       (2, 2)),
  )
  def test_ray_exception_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(perspective.ray, error_msg, shapes)

  @parameterized.parameters(
      (((0., 0.), (1., 1.), (0., 0.)), ((0., 0., 1.),)),
      (((0., 0.), (1., 1.), (-1., -2.)), ((1., 2., 1.),)),
      (((0., 0.), (10., 1.), (-1., -2.)), ((.1, 2., 1.),)),
      (((-2., -4.), (10., 1.), (-3., -6.)), ((.1, 2., 1.),)),
  )
  def test_ray_preset(self, test_inputs, test_outputs):
    """Tests that the ray function gives the correct result."""
    self.assert_output_is_correct(perspective.ray, test_inputs, test_outputs)

  def test_ray_project_random(self):
    """Tests that the end point of the ray projects at the good location."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    random_point_2d = np.random.normal(size=tensor_shape + [2])
    random_focal = np.random.normal(size=tensor_shape + [2])
    random_principal_point = np.random.normal(size=tensor_shape + [2])

    ray_3d = perspective.ray(random_point_2d, random_focal,
                             random_principal_point)
    point_2d = perspective.project(ray_3d, random_focal, random_principal_point)

    self.assertAllClose(random_point_2d, point_2d, rtol=1e-3)

  @parameterized.parameters(
      ((2,), (1,), (2,), (2,)),
      ((2, 2), (2, 1), (2, 2), (2, 2)),
      ((None, 2), (None, 1), (None, 2), (None, 2)),
  )
  def test_unproject_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(perspective.unproject, shapes)

  @parameterized.parameters(
      ("must have exactly 2 dimensions in axis -1", (None,), (1,), (2,), (2,)),
      ("must have exactly 1 dimensions in axis -1", (2,), (None,), (2,), (2,)),
      ("must have exactly 2 dimensions in axis -1", (2,), (1,), (None,), (2,)),
      ("must have exactly 2 dimensions in axis -1", (2,), (1,), (2,), (None,)),
      ("Not all batch dimensions are identical.", (1, 2), (2, 1), (2, 2),
       (2, 2)),
  )
  def test_unproject_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(perspective.unproject, error_msg, shapes)

  @parameterized.parameters(
      (((0., 0.), (1.,), (1., 1.), (0., 0.)), ((0., 0., 1.),)),
      (((0., 0.), (1.,), (1., 1.), (-4., -2.)), ((4., 2., 1.),)),
      (((0., 0.), (10.,), (1., 1.), (-.4, -.2)), ((4., 2., 10.),)),
      (((0., 0.), (10.,), (2., 1.), (-.8, -.2)), ((4., 2., 10.),)),
      (((0., .2), (10.,), (2., 1.), (-.8, 0.)), ((4., 2., 10.),)),
  )
  def test_unproject_preset(self, test_inputs, test_outputs):
    """Tests that the unproject function gives the correct result."""
    self.assert_output_is_correct(perspective.unproject, test_inputs,
                                  test_outputs)

  def test_unproject_project_random(self):
    """Tests that unprojecting and projecting gives and identity mapping."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    random_point_2d = np.random.normal(size=tensor_shape + [2])
    random_focal = np.random.normal(size=tensor_shape + [2])
    random_principal_point = np.random.normal(size=tensor_shape + [2])
    random_depth = np.random.normal(size=tensor_shape + [1])

    point_3d = perspective.unproject(random_point_2d, random_depth,
                                     random_focal, random_principal_point)
    point_2d = perspective.project(point_3d, random_focal,
                                   random_principal_point)

    self.assertAllClose(random_point_2d, point_2d, rtol=1e-3)

  def test_unproject_ray_random(self):
    """Tests that that ray is pointing toward the correct location."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    random_point_2d = np.random.normal(size=tensor_shape + [2])
    random_focal = np.random.normal(size=tensor_shape + [2])
    random_principal_point = np.random.normal(size=tensor_shape + [2])
    random_depth = np.random.normal(size=tensor_shape + [1])

    point_3d = perspective.unproject(random_point_2d, random_depth,
                                     random_focal, random_principal_point)
    ray_3d = perspective.ray(random_point_2d, random_focal,
                             random_principal_point)
    ray_3d = random_depth * ray_3d

    self.assertAllClose(point_3d, ray_3d, rtol=1e-3)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""ModelNet40 classification dataset from https://modelnet.cs.princeton.edu."""

import os
import h5py
import tensorflow as tf
import tensorflow_datasets as tfds

# Constants
_URL = 'https://shapenet.cs.stanford.edu/media/modelnet40_ply_hdf5_2048.zip'

_CITATION = """
@inproceedings{qi2017pointnet,
  title={Pointnet: Deep learning on point sets for 3d classification and segmentation},
  author={Qi, Charles R and Su, Hao and Mo, Kaichun and Guibas, Leonidas J},
  booktitle={Proceedings of the IEEE conference on computer vision and pattern recognition},
  pages={652--660},
  year={2017}
}
"""

_DESCRIPTION = """
The dataset contains point clouds sampling CAD models from 40 different categories.
The files have been retrieved from https://modelnet.cs.princeton.edu

To generate each example, the authors first uniformly sampled a modelnet40 mesh
with 10000 uniform samples, and then employed furthest point sampling to downsample
to 2048 points. The procedure is explained [here](https://github.com/charlesq34/pointnet2/blob/master/tf_ops/sampling/tf_sampling.py)
"""

_LABELS = [
    'airplane', 'bathtub', 'bed', 'bench', 'bookshelf', 'bottle', 'bowl', 'car',
    'chair', 'cone', 'cup', 'curtain', 'desk', 'door', 'dresser', 'flower_pot',
    'glass_box', 'guitar', 'keyboard', 'lamp', 'laptop', 'mantel', 'monitor',
    'night_stand', 'person', 'piano', 'plant', 'radio', 'range_hood', 'sink',
    'sofa', 'stairs', 'stool', 'table', 'tent', 'toilet', 'tv_stand', 'vase',
    'wardrobe', 'xbox'
]

# --- registers the checksum
_CHECKSUM_DIR = os.path.join(os.path.dirname(__file__), 'checksums/')
_CHECKSUM_DIR = os.path.normpath(_CHECKSUM_DIR)
tfds.download.add_checksums_dir(_CHECKSUM_DIR)


class ModelNet40(tfds.core.GeneratorBasedBuilder):
  """ModelNet40."""

  VERSION = tfds.core.Version('1.0.0')

  @staticmethod
  def load(*args, **kwargs):
    return tfds.load('model_net40', *args, **kwargs)

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'points': tfds.features.Tensor(shape=(2048, 3), dtype=tf.float32),
            'label': tfds.features.ClassLabel(names=_LABELS)
        }),
        supervised_keys=('points', 'label'),
        homepage='https://modelnet.cs.princeton.edu',
        citation=_CITATION,
    )

  def _split_generators(self, download_manager):
    """Returns SplitGenerators."""

    extracted_path = download_manager.download_and_extract(_URL)

    # Note: VALIDATION split was not provided by the authors
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs=dict(
                filename_list_path=os.path.join(
                    extracted_path,
                    'modelnet40_ply_hdf5_2048/train_files.txt'),)),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs=dict(
                filename_list_path=os.path.join(
                    extracted_path,
                    'modelnet40_ply_hdf5_2048/test_files.txt'),)),
    ]

  def _generate_examples(self, filename_list_path):
    """Yields examples."""

    ancestor_path = os.path.dirname(os.path.dirname(filename_list_path))
    with tf.io.gfile.GFile(filename_list_path, 'r') as fid:
      filename_list = fid.readlines()
    filename_list = [line.rstrip()[5:] for line in filename_list]

    example_key = -1  # as yield exists, need to pre-increment
    for filename in filename_list:
      h5path = os.path.join(ancestor_path, filename)

      with h5py.File(h5path, 'r') as h5file:
        points = h5file['data'][:]  # shape=(2048, 2048, 3)
        label = h5file['label'][:]  # shape=(2048, )

        models_per_file = points.shape[0]
        for imodel in range(models_per_file):
          example_key += 1
          yield example_key, {
              'points': points[imodel, :, :],
              'label': int(label[imodel])
          }



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OpenGL module."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow_graphics.rendering.opengl import math
# # google internal import 1
from tensorflow_graphics.util import export_api as _export_api

# pylint: disable=g-import-not-at-top
try:
  from tensorflow_graphics.rendering.opengl import gen_rasterizer_op as rasterizer
except ImportError:
  import tensorflow as tf
  rasterizer = tf.load_op_library(
      tf.compat.v1.resource_loader.get_path_to_datafile('rasterizer_op.so'))
# pylint: enable=g-import-not-at-top

# API contains submodules of tensorflow_graphics.rendering.
__all__ = _export_api.get_modules()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements image matting functionalities."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf

from tensorflow_graphics.math import vector as tfg_vector
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape


def _shape(batch_shape, *shapes):
  """Creates a new shape concatenating batch_shape and shapes.

  Args:
    batch_shape: A Tensor or list/tuple containing the batch dimensions.
    *shapes: A tuple containing new dimensions to append to the final shape.

  Returns:
    A Tensor containing the final shape dimensions.
  """
  return tf.concat((batch_shape, shapes), axis=-1)


def _quadratic_form(matrix, vector):
  """Computes the quadratic form between a matrix and a vector.

  The quadratic form between a matrix A and a vector x can be written as
  Q(x) = <x,Ax>, where <.,.> is the dot product operator.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    matrix: A tensor of shape `[A1, ..., An, C, C]`.
    vector: A tensor of shape `[A1, ..., An, 1, C]`.

  Returns:
    A tensor of shape `[A1, ..., An, 1]`.
  """
  vector_matrix = tf.matmul(vector, matrix)
  vector_matrix_vector = tf.matmul(vector_matrix, vector, transpose_b=True)
  return vector_matrix_vector


def _image_patches(image, size):
  """Extracts square image patches.

  Args:
    image: A tensor of shape `[B, H, W, C]`.
    size: The size of the square patches.

  Returns:
    A tensor of shape `[B, H - pad, W - pad, C * size^2]`, where `pad` is
    `size - 1`.
  """
  return tf.image.extract_patches(
      image,
      sizes=(1, size, size, 1),
      strides=(1, 1, 1, 1),
      rates=(1, 1, 1, 1),
      padding="VALID")


def _image_average(image, size):
  """Computes average over image patches.

  Args:
    image: A tensor of shape `[B, H, W, C]`.
    size: The size of the square patches.

  Returns:
    A tensor of shape `[B, H - pad, W - pad, C]`, where `pad` is `size - 1`.
  """
  return tf.compat.v1.nn.avg_pool2d(
      value=image,
      ksize=(1, size, size, 1),
      strides=(1, 1, 1, 1),
      padding="VALID")


def build_matrices(image, size=3, eps=1e-5, name=None):
  """Generates the closed form matting Laplacian.

  Generates the closed form matting Laplacian as proposed by Levin et
  al. in "A Closed Form Solution to Natural Image Matting". This function also
  return the pseudo-inverse matrix allowing to retrieve the matting linear
  coefficient.

  Args:
    image: A tensor of shape `[B, H, W, C]`.
    size: An `int` representing the size of the patches used to enforce
      smoothness.
    eps: A small number of type `float` to regularize the problem.
    name: A name for this op. Defaults to "matting_build_matrices".

  Returns:
    A tensor of shape `[B, H - pad, W - pad, size^2, size^2]` containing
    the matting Laplacian matrices. A tensor of shape
    `[B, H - pad, W - pad, C + 1, size^2]` containing the pseudo-inverse
    matrices which can be used to retrieve the matting linear coefficients.
    The padding `pad` is equal to `size - 1`.

  Raises:
    ValueError: If `image` is not of rank 4.
  """
  with tf.compat.v1.name_scope(name, "matting_build_matrices", [image]):
    image = tf.convert_to_tensor(value=image)
    eps = tf.constant(value=eps, dtype=image.dtype)

    shape.check_static(image, has_rank=4)
    if size % 2 == 0:
      raise ValueError("The patch size is expected to be an odd value.")

    pixels = size**2
    channels = tf.shape(input=image)[-1]
    dtype = image.dtype
    # Extracts image patches.
    patches = _image_patches(image, size)
    batches = tf.shape(input=patches)[:-1]
    patches = tf.reshape(patches, shape=_shape(batches, pixels, channels))
    # Creates the data matrix block.
    ones = tf.ones(shape=_shape(batches, pixels, 1), dtype=dtype)
    affine = tf.concat((patches, ones), axis=-1)
    # Creates the regularizer matrix block.
    diag = tf.sqrt(eps) * tf.eye(channels, batch_shape=(1, 1, 1), dtype=dtype)
    zeros = tf.zeros(shape=_shape((1, 1, 1), channels, 1), dtype=dtype)
    regularizer = tf.concat((diag, zeros), axis=-1)
    regularizer = tf.tile(regularizer, multiples=_shape(batches, 1, 1))
    # Creates a matrix concatenating the data and regularizer blocks.
    mat = tf.concat((affine, regularizer), axis=-2)
    # Builds the pseudo inverse and the laplacian matrices.
    inverse = tf.linalg.inv(tf.matmul(mat, mat, transpose_a=True))
    inverse = asserts.assert_no_infs_or_nans(inverse)
    pseudo_inverse = tf.matmul(inverse, affine, transpose_b=True)
    identity = tf.eye(num_rows=pixels, dtype=dtype)
    laplacian = identity - tf.matmul(affine, pseudo_inverse)
    return laplacian, pseudo_inverse


def linear_coefficients(matte, pseudo_inverse, name=None):
  """Computes the matting linear coefficients.

  Computes the matting linear coefficients (a, b) based on the `pseudo_inverse`
  generated by the `build_matrices` function which implements the approach
  proposed by Levin et al. in "A Closed Form Solution to Natural Image Matting".

  Args:
    matte: A tensor of shape `[B, H, W, 1]`.
    pseudo_inverse: A tensor of shape `[B, H - pad, W - pad, C + 1, size^2]`
      containing the pseudo-inverse matrices computed by the `build_matrices`
      function, where `pad` is equal to `size - 1` and `size` is the patch size
      used to compute this tensor.
    name: A name for this op. Defaults to "matting_linear_coefficients".

  Returns:
    A tuple contraining two Tensors for the linear coefficients (a, b) of shape
    `[B, H, W, C]` and `[B, H, W, 1]`.

  Raises:
    ValueError: If the last dimension of `matte` is not 1. If `matte` is not
    of rank 4. If `pseudo_inverse` is not of rank 5. If `B` is different
    between `matte` and `pseudo_inverse`.
  """
  with tf.compat.v1.name_scope(name, "matting_linear_coefficients",
                               [matte, pseudo_inverse]):
    matte = tf.convert_to_tensor(value=matte)
    pseudo_inverse = tf.convert_to_tensor(value=pseudo_inverse)

    pixels = tf.compat.dimension_value(pseudo_inverse.shape[-1])
    shape.check_static(matte, has_rank=4, has_dim_equals=(-1, 1))
    shape.check_static(pseudo_inverse, has_rank=5)
    shape.compare_batch_dimensions(
        tensors=(matte, pseudo_inverse),
        last_axes=0,
        broadcast_compatible=False)

    size = np.sqrt(pixels)
    # Computes the linear coefficients.
    patches = tf.expand_dims(_image_patches(matte, size), axis=-1)
    coeffs = tf.squeeze(tf.matmul(pseudo_inverse, patches), axis=-1)
    # Averages the linear coefficients over patches.
    height = tf.shape(input=coeffs)[1]
    width = tf.shape(input=coeffs)[2]
    ones = tf.ones(shape=_shape((1,), height, width, 1), dtype=matte.dtype)
    height = tf.shape(input=matte)[1] + size - 1
    width = tf.shape(input=matte)[2] + size - 1
    coeffs = tf.image.resize_with_crop_or_pad(coeffs, height, width)
    ones = tf.image.resize_with_crop_or_pad(ones, height, width)
    coeffs = _image_average(coeffs, size) / _image_average(ones, size)
    return tf.split(coeffs, (-1, 1), axis=-1)


def loss(matte, laplacian, name=None):
  """Computes the matting loss function based on the matting Laplacian.

  Computes the matting loss function based on the `laplacian` generated by the
  `build_matrices` function which implements the approach proposed by Levin
  et al. in "A Closed Form Solution to Natural Image Matting".

  Args:
    matte: A tensor of shape `[B, H, W, 1]`.
    laplacian: A tensor of shape `[B, H - pad, W - pad, size^2, size^2]`
      containing the Laplacian matrices computed by the `build_matrices`
      function, where `pad` is equal to `size - 1` and `size` is the patch size
      used to compute this tensor.
    name: A name for this op. Defaults to "matting_loss".

  Returns:
    A tensor containing a scalar value defining the matting loss.

  Raises:
    ValueError: If the last dimension of `matte` is not 1. If `matte` is not
    of rank 4. If the last two dimensions of `laplacian` are not of the
    same size. If `laplacian` is not of rank 5. If `B` is different
    between `matte` and `laplacian`.
  """
  with tf.compat.v1.name_scope(name, "matting_loss", [matte, laplacian]):
    matte = tf.convert_to_tensor(value=matte)
    laplacian = tf.convert_to_tensor(value=laplacian)

    pixels = tf.compat.dimension_value(laplacian.shape[-1])
    shape.check_static(matte, has_rank=4, has_dim_equals=(-1, 1))
    shape.check_static(laplacian, has_rank=5, has_dim_equals=(-2, pixels))
    shape.compare_batch_dimensions(
        tensors=(matte, laplacian), last_axes=0, broadcast_compatible=False)

    size = np.sqrt(pixels)
    patches = tf.expand_dims(_image_patches(matte, size), axis=-2)
    losses = _quadratic_form(laplacian, patches)
    return tf.reduce_mean(input_tensor=losses)


def reconstruct(image, coeff_mul, coeff_add, name=None):
  """Reconstruct the matte from the image using the linear coefficients.

  Reconstruct the matte from the image using the linear coefficients (a, b)
  returned by the linear_coefficients function.

  Args:
    image: A tensor of shape `[B, H, W, C]` .
    coeff_mul: A tensor of shape `[B, H, W, C]` representing the multiplicative
      part of the linear coefficients.
    coeff_add: A tensor of shape `[B, H, W, 1]` representing the additive part
      of the linear coefficients.
    name: A name for this op. Defaults to "matting_reconstruct".

  Returns:
    A tensor of shape `[B, H, W, 1]` containing the mattes.

  Raises:
    ValueError: If `image`, `coeff_mul`, or `coeff_add` are not of rank 4. If
    the last dimension of `coeff_add` is not 1. If the batch dimensions of
    `image`, `coeff_mul`, and `coeff_add` do not match.
  """
  with tf.compat.v1.name_scope(name, "matting_reconstruct",
                               [image, coeff_mul, coeff_add]):
    image = tf.convert_to_tensor(value=image)
    coeff_mul = tf.convert_to_tensor(value=coeff_mul)
    coeff_add = tf.convert_to_tensor(value=coeff_add)

    shape.check_static(image, has_rank=4)
    shape.check_static(coeff_mul, has_rank=4)
    shape.check_static(coeff_add, has_rank=4, has_dim_equals=(-1, 1))
    shape.compare_batch_dimensions(
        tensors=(image, coeff_mul), last_axes=-1, broadcast_compatible=False)
    shape.compare_batch_dimensions(
        tensors=(image, coeff_add), last_axes=-2, broadcast_compatible=False)

    return tfg_vector.dot(coeff_mul, image) + coeff_add


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for export API functionalities."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import test_case


class ExportAPITest(test_case.TestCase):

  def test_get_functions_and_classes(self):
    """Tests that get_functions_and_classes does not raise an exception."""
    try:
      export_api.get_functions_and_classes()
    except Exception as e:  # pylint: disable=broad-except
      self.fail("Exception raised: %s" % str(e))

  def test_get_modules(self):
    """Tests that get_modules does not raise an exception."""
    try:
      export_api.get_modules()
    except Exception as e:  # pylint: disable=broad-except
      self.fail("Exception raised: %s" % str(e))


if __name__ == "__main__":
  test_case.main()
